# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from tests import ClientHTTPStubber, create_session, mock

OPERATION_PARAMS = {
    'change_password': {
        'PreviousPassword': 'myoldbadpassword',
        'ProposedPassword': 'mynewgoodpassword',
        'AccessToken': 'foobar',
    },
    'confirm_forgot_password': {
        'ClientId': 'foo',
        'Username': 'myusername',
        'ConfirmationCode': 'thisismeforreal',
        'Password': 'whydowesendpasswordsviaemail',
    },
    'confirm_sign_up': {
        'ClientId': 'foo',
        'Username': 'myusername',
        'ConfirmationCode': 'ireallydowanttosignup',
    },
    'delete_user': {'AccessToken': 'foobar'},
    'delete_user_attributes': {
        'UserAttributeNames': ['myattribute'],
        'AccessToken': 'foobar',
    },
    'forgot_password': {'ClientId': 'foo', 'Username': 'myusername'},
    'get_user': {'AccessToken': 'foobar'},
    'get_user_attribute_verification_code': {
        'AttributeName': 'myattribute',
        'AccessToken': 'foobar',
    },
    'resend_confirmation_code': {'ClientId': 'foo', 'Username': 'myusername'},
    'set_user_settings': {
        'AccessToken': 'randomtoken',
        'MFAOptions': [
            {'DeliveryMedium': 'SMS', 'AttributeName': 'someattributename'}
        ],
    },
    'sign_up': {
        'ClientId': 'foo',
        'Username': 'bar',
        'Password': 'mysupersecurepassword',
    },
    'update_user_attributes': {
        'UserAttributes': [{'Name': 'someattributename', 'Value': 'newvalue'}],
        'AccessToken': 'foobar',
    },
    'verify_user_attribute': {
        'AttributeName': 'someattributename',
        'Code': 'someverificationcode',
        'AccessToken': 'foobar',
    },
}


@pytest.mark.parametrize(
    "operation_name, parameters", OPERATION_PARAMS.items()
)
def test_unsigned_operations(operation_name, parameters):
    environ = {
        'AWS_ACCESS_KEY_ID': 'access_key',
        'AWS_SECRET_ACCESS_KEY': 'secret_key',
        'AWS_CONFIG_FILE': 'no-exist-foo',
    }

    with mock.patch('os.environ', environ):
        session = create_session()
        session.config_filename = 'no-exist-foo'
        client = session.create_client('cognito-idp', 'us-west-2')
        http_stubber = ClientHTTPStubber(client)

        operation = getattr(client, operation_name)

        http_stubber.add_response(body=b'{}')
        with http_stubber:
            operation(**parameters)
            request = http_stubber.requests[0]

        assert 'authorization' not in request.headers, (
            'authorization header found in unsigned operation'
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import re
import shutil
import tempfile

from botocore.docs.service import ServiceDocumenter
from botocore.session import get_session
from tests import unittest


class BaseDocsFunctionalTest(unittest.TestCase):
    def setUp(self):
        self._session = get_session()
        self.docs_root_dir = tempfile.mkdtemp()
        self.root_services_path = os.path.join(
            self.docs_root_dir, 'reference', 'services'
        )

    def tearDown(self):
        shutil.rmtree(self.docs_root_dir)

    def assert_contains_line(self, line, contents):
        contents = contents.decode('utf-8')
        self.assertIn(line, contents)

    def assert_contains_lines_in_order(self, lines, contents):
        contents = contents.decode('utf-8')
        for line in lines:
            self.assertIn(line, contents)
            beginning = contents.find(line)
            contents = contents[(beginning + len(line)) :]

    def assert_not_contains_line(self, line, contents):
        contents = contents.decode('utf-8')
        self.assertNotIn(line, contents)

    def assert_not_contains_lines(self, lines, contents):
        contents = contents.decode('utf-8')
        for line in lines:
            self.assertNotIn(line, contents)

    def get_client_method_contents(self, service_name, method_name):
        service_file_path = os.path.join(
            self.root_services_path,
            service_name,
            'client',
            f'{method_name}.rst',
        )
        with open(service_file_path, 'rb') as f:
            return f.read()

    def get_title_section_for(self, service_name):
        contents = (
            ServiceDocumenter(
                service_name, self._session, self.root_services_path
            )
            .document_service()
            .decode('utf-8')
        )
        start_of_table_of_contents = 'Table of Contents'
        start_index = contents.find(start_of_table_of_contents)
        contents = contents[:start_index]
        contents = contents.encode('utf-8')
        return contents

    def get_method_document_block(self, operation_name, contents):
        contents = contents.decode('utf-8')
        regex = rf'.. py:method:: ([a-zA-Z0-9]*\.)*{operation_name}\('
        match = re.search(regex, contents)
        self.assertIsNotNone(match, 'Method is not found in contents')
        start_method_document = match.group()
        start_index = match.start()
        contents = contents[start_index:]
        end_index = contents.find('.. py:method::', len(start_method_document))
        contents = contents[:end_index]
        return contents.encode('utf-8')

    def get_parameter_document_block(self, param_name, contents):
        contents = contents.decode('utf-8')
        start_param_document = f'  :type {param_name}:'
        start_index = contents.find(start_param_document)
        self.assertNotEqual(start_index, -1, 'Param is not found in contents')
        contents = contents[start_index:]
        end_index = contents.find('  :type', len(start_param_document))
        contents = contents[:end_index]
        return contents.encode('utf-8')

    def get_parameter_documentation_from_service(
        self, service_name, method_name, param_name
    ):
        ServiceDocumenter(
            service_name, self._session, self.root_services_path
        ).document_service()
        contents = self.get_client_method_contents(service_name, method_name)
        method_contents = self.get_method_document_block(method_name, contents)
        return self.get_parameter_document_block(param_name, method_contents)

    def get_docstring_for_method(self, service_name, method_name):
        ServiceDocumenter(
            service_name, self._session, self.root_services_path
        ).document_service()
        contents = self.get_client_method_contents(service_name, method_name)
        method_contents = self.get_method_document_block(method_name, contents)
        return method_contents

    def assert_is_documented_as_autopopulated_param(
        self, service_name, method_name, param_name, doc_string=None
    ):
        ServiceDocumenter(
            service_name, self._session, self.root_services_path
        ).document_service()
        contents = self.get_client_method_contents(service_name, method_name)
        method_contents = self.get_method_document_block(method_name, contents)

        # Ensure it is not in the example.
        self.assert_not_contains_line(
            f'{param_name}=\'string\'', method_contents
        )

        # Ensure it is in the params.
        param_contents = self.get_parameter_document_block(
            param_name, method_contents
        )

        # Ensure it is not labeled as required.
        self.assert_not_contains_line('REQUIRED', param_contents)

        # Ensure the note about autopopulation was added.
        if doc_string is None:
            doc_string = 'Please note that this parameter is automatically'
        self.assert_contains_line(doc_string, param_contents)



# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import binascii
import contextlib
import datetime
import os
import platform
import random
import select
import shutil
import sys
import tempfile
import time
import unittest
from contextlib import ContextDecorator
from io import BytesIO
from subprocess import PIPE, Popen
from unittest import mock

from dateutil.tz import tzlocal

import botocore.loaders
import botocore.session
from botocore import credentials, utils
from botocore.awsrequest import AWSResponse
from botocore.compat import HAS_CRT, parse_qs, urlparse
from botocore.configprovider import create_botocore_default_config_mapping
from botocore.httpchecksum import _CHECKSUM_CLS, DEFAULT_CHECKSUM_ALGORITHM
from botocore.stub import Stubber

_LOADER = botocore.loaders.Loader()


def _all_services():
    session = botocore.session.Session()
    service_names = session.get_available_services()
    return [session.get_service_model(name) for name in service_names]


# Only compute our service models once
ALL_SERVICES = _all_services()


def skip_unless_has_memory_collection(cls):
    """Class decorator to skip tests that require memory collection.

    Any test that uses memory collection (such as the resource leak tests)
    can decorate their class with skip_unless_has_memory_collection to
    indicate that if the platform does not support memory collection
    the tests should be skipped.
    """
    if platform.system() not in ['Darwin', 'Linux']:
        return unittest.skip('Memory tests only supported on mac/linux.')(cls)
    return cls


def skip_if_windows(reason):
    """Decorator to skip tests that should not be run on windows.
    Example usage:
        @skip_if_windows("Not valid")
        def test_some_non_windows_stuff(self):
            self.assertEqual(...)
    """

    def decorator(func):
        return unittest.skipIf(
            platform.system() not in ['Darwin', 'Linux'], reason
        )(func)

    return decorator


def requires_crt(reason=None):
    if reason is None:
        reason = "Test requires awscrt to be installed"

    def decorator(func):
        return unittest.skipIf(not HAS_CRT, reason)(func)

    return decorator


def random_chars(num_chars):
    """Returns random hex characters.

    Useful for creating resources with random names.

    """
    return binascii.hexlify(os.urandom(int(num_chars / 2))).decode('ascii')


def create_session(**kwargs):
    # Create a Session object.  By default,
    # the _LOADER object is used as the loader
    # so that we reused the same models across tests.
    session = botocore.session.Session(**kwargs)
    session.register_component('data_loader', _LOADER)
    session.set_config_variable('credentials_file', 'noexist/foo/botocore')
    return session


def get_botocore_default_config_mapping():
    session = botocore.session.get_session()
    return create_botocore_default_config_mapping(session)


@contextlib.contextmanager
def temporary_file(mode):
    """This is a cross platform temporary file creation.

    tempfile.NamedTemporary file on windows creates a secure temp file
    that can't be read by other processes and can't be opened a second time.

    For tests, we generally *want* them to be read multiple times.
    The test fixture writes the temp file contents, the test reads the
    temp file.

    """
    temporary_directory = tempfile.mkdtemp()
    basename = f'tmpfile-{int(time.time())}-{random.randint(1, 1000)}'
    full_filename = os.path.join(temporary_directory, basename)
    open(full_filename, 'w').close()
    try:
        with open(full_filename, mode) as f:
            yield f
    finally:
        shutil.rmtree(temporary_directory)


class BaseEnvVar(unittest.TestCase):
    def setUp(self):
        # Automatically patches out os.environ for you
        # and gives you a self.environ attribute that simulates
        # the environment.  Also will automatically restore state
        # for you in tearDown()
        self.environ = {}
        self.environ_patch = mock.patch('os.environ', self.environ)
        self.environ_patch.start()

    def tearDown(self):
        self.environ_patch.stop()


class BaseSessionTest(BaseEnvVar):
    """Base class used to provide credentials.

    This class can be used as a base class that want to use a real
    session class but want to be completely isolated from the
    external environment (including environment variables).

    This class will also set credential vars so you can make fake
    requests to services.

    """

    def setUp(self, **environ):
        super().setUp()
        self.environ['AWS_ACCESS_KEY_ID'] = 'access_key'
        self.environ['AWS_SECRET_ACCESS_KEY'] = 'secret_key'
        self.environ['AWS_CONFIG_FILE'] = 'no-exist-foo'
        self.environ.update(environ)
        self.session = create_session()
        self.session.config_filename = 'no-exist-foo'


@skip_unless_has_memory_collection
class BaseClientDriverTest(unittest.TestCase):
    INJECT_DUMMY_CREDS = False

    def setUp(self):
        self.driver = ClientDriver()
        env = None
        if self.INJECT_DUMMY_CREDS:
            env = {'AWS_ACCESS_KEY_ID': 'foo', 'AWS_SECRET_ACCESS_KEY': 'bar'}
        self.driver.start(env=env)

    def cmd(self, *args):
        self.driver.cmd(*args)

    def send_cmd(self, *args):
        self.driver.send_cmd(*args)

    def record_memory(self):
        self.driver.record_memory()

    @property
    def memory_samples(self):
        return self.driver.memory_samples

    def tearDown(self):
        self.driver.stop()


class ClientDriver:
    CLIENT_SERVER = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'cmd-runner'
    )

    def __init__(self):
        self._popen = None
        self.memory_samples = []

    def _get_memory_with_ps(self, pid):
        # It would be better to eventually switch to psutil,
        # which should allow us to test on windows, but for now
        # we'll just use ps and run on POSIX platforms.
        command_list = ['ps', '-p', str(pid), '-o', 'rss']
        p = Popen(command_list, stdout=PIPE)
        stdout = p.communicate()[0]
        if not p.returncode == 0:
            raise RuntimeError("Could not retrieve memory")
        else:
            # Get the RSS from output that looks like this:
            # RSS
            # 4496
            return int(stdout.splitlines()[1].split()[0]) * 1024

    def record_memory(self):
        mem = self._get_memory_with_ps(self._popen.pid)
        self.memory_samples.append(mem)

    def start(self, env=None):
        """Start up the command runner process."""
        self._popen = Popen(
            [sys.executable, self.CLIENT_SERVER],
            stdout=PIPE,
            stdin=PIPE,
            env=env,
        )

    def stop(self):
        """Shutdown the command runner process."""
        self.cmd('exit')
        self._popen.wait()

    def send_cmd(self, *cmd):
        """Send a command and return immediately.

        This is a lower level method than cmd().
        This method will instruct the cmd-runner process
        to execute a command, but this method will
        immediately return.  You will need to use
        ``is_cmd_finished()`` to check that the command
        is finished.

        This method is useful if you want to record attributes
        about the process while an operation is occurring.  For
        example, if you want to instruct the cmd-runner process
        to upload a 1GB file to S3 and you'd like to record
        the memory during the upload process, you can use
        send_cmd() instead of cmd().

        """
        cmd_str = ' '.join(cmd) + '\n'
        cmd_bytes = cmd_str.encode('utf-8')
        self._popen.stdin.write(cmd_bytes)
        self._popen.stdin.flush()

    def is_cmd_finished(self):
        rlist = [self._popen.stdout.fileno()]
        result = select.select(rlist, [], [], 0.01)
        if result[0]:
            return True
        return False

    def cmd(self, *cmd):
        """Send a command and block until it finishes.

        This method will send a command to the cmd-runner process
        to run.  It will block until the cmd-runner process is
        finished executing the command and sends back a status
        response.

        """
        self.send_cmd(*cmd)
        result = self._popen.stdout.readline().strip()
        if result != b'OK':
            raise RuntimeError(f"Error from command '{cmd}': {result}")


# This is added to this file because it's used in both
# the functional and unit tests for cred refresh.
class IntegerRefresher(credentials.RefreshableCredentials):
    """Refreshable credentials to help with testing.

    This class makes testing refreshable credentials easier.
    It has the following functionality:

        * A counter, self.refresh_counter, to indicate how many
          times refresh was called.
        * A way to specify how many seconds to make credentials
          valid.
        * Configurable advisory/mandatory refresh.
        * An easy way to check consistency.  Each time creds are
          refreshed, all the cred values are set to the next
          incrementing integer.  Frozen credentials should always
          have this value.
    """

    _advisory_refresh_timeout = 2
    _mandatory_refresh_timeout = 1
    _credentials_expire = 3

    def __init__(
        self,
        creds_last_for=_credentials_expire,
        advisory_refresh=_advisory_refresh_timeout,
        mandatory_refresh=_mandatory_refresh_timeout,
        refresh_function=None,
    ):
        expires_in = self._current_datetime() + datetime.timedelta(
            seconds=creds_last_for
        )
        if refresh_function is None:
            refresh_function = self._do_refresh
        super().__init__(
            '0', '0', '0', expires_in, refresh_function, 'INTREFRESH'
        )
        self.creds_last_for = creds_last_for
        self.refresh_counter = 0
        self._advisory_refresh_timeout = advisory_refresh
        self._mandatory_refresh_timeout = mandatory_refresh

    def _do_refresh(self):
        self.refresh_counter += 1
        current = int(self._access_key)
        next_id = str(current + 1)

        return {
            'access_key': next_id,
            'secret_key': next_id,
            'token': next_id,
            'expiry_time': self._seconds_later(self.creds_last_for),
        }

    def _seconds_later(self, num_seconds):
        # We need to guarantee at *least* num_seconds.
        # Because this doesn't handle subsecond precision
        # we'll round up to the next second.
        num_seconds += 1
        t = self._current_datetime() + datetime.timedelta(seconds=num_seconds)
        return self._to_timestamp(t)

    def _to_timestamp(self, datetime_obj):
        obj = utils.parse_to_aware_datetime(datetime_obj)
        return obj.strftime('%Y-%m-%dT%H:%M:%SZ')

    def _current_timestamp(self):
        return self._to_timestamp(self._current_datetime())

    def _current_datetime(self):
        return datetime.datetime.now(tzlocal())


def _urlparse(url):
    if isinstance(url, bytes):
        # Not really necessary, but it helps to reduce noise on Python 2.x
        url = url.decode('utf8')
    return urlparse(url)


def assert_url_equal(url1, url2):
    parts1 = _urlparse(url1)
    parts2 = _urlparse(url2)

    # Because the query string ordering isn't relevant, we have to parse
    # every single part manually and then handle the query string.
    assert parts1.scheme == parts2.scheme
    assert parts1.netloc == parts2.netloc
    assert parts1.path == parts2.path
    assert parts1.params == parts2.params
    assert parts1.fragment == parts2.fragment
    assert parts1.username == parts2.username
    assert parts1.password == parts2.password
    assert parts1.hostname == parts2.hostname
    assert parts1.port == parts2.port
    assert parse_qs(parts1.query) == parse_qs(parts2.query)


class HTTPStubberException(Exception):
    pass


class RawResponse(BytesIO):
    # TODO: There's a few objects similar to this in various tests, let's
    # try and consolidate to this one in a future commit.
    def stream(self, **kwargs):
        contents = self.read()
        while contents:
            yield contents
            contents = self.read()


class BaseHTTPStubber:
    def __init__(self, obj_with_event_emitter, strict=True):
        self.reset()
        self._strict = strict
        self._obj_with_event_emitter = obj_with_event_emitter

    def reset(self):
        self.requests = []
        self.responses = []

    def add_response(
        self, url='https://example.com', status=200, headers=None, body=b''
    ):
        if headers is None:
            headers = {}

        raw = RawResponse(body)
        response = AWSResponse(url, status, headers, raw)
        self.responses.append(response)

    @property
    def _events(self):
        raise NotImplementedError('_events')

    def start(self):
        self._events.register('before-send', self)

    def stop(self):
        self._events.unregister('before-send', self)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __call__(self, request, **kwargs):
        self.requests.append(request)
        if self.responses:
            response = self.responses.pop(0)
            if isinstance(response, Exception):
                raise response
            else:
                return response
        elif self._strict:
            raise HTTPStubberException('Insufficient responses')
        else:
            return None


class ClientHTTPStubber(BaseHTTPStubber):
    @property
    def _events(self):
        return self._obj_with_event_emitter.meta.events


class SessionHTTPStubber(BaseHTTPStubber):
    @property
    def _events(self):
        return self._obj_with_event_emitter.get_component('event_emitter')


class ConsistencyWaiterException(Exception):
    pass


class ConsistencyWaiter:
    """
    A waiter class for some check to reach a consistent state.

    :type min_successes: int
    :param min_successes: The minimum number of successful check calls to
    treat the check as stable. Default of 1 success.

    :type max_attempts: int
    :param min_successes: The maximum number of times to attempt calling
    the check. Default of 20 attempts.

    :type delay: int
    :param delay: The number of seconds to delay the next API call after a
    failed check call. Default of 5 seconds.
    """

    def __init__(
        self,
        min_successes=1,
        max_attempts=20,
        delay=5,
        delay_initial_poll=False,
    ):
        self.min_successes = min_successes
        self.max_attempts = max_attempts
        self.delay = delay
        self.delay_initial_poll = delay_initial_poll

    def wait(self, check, *args, **kwargs):
        """
        Wait until the check succeeds the configured number of times

        :type check: callable
        :param check: A callable that returns True or False to indicate
        if the check succeeded or failed.

        :type args: list
        :param args: Any ordered arguments to be passed to the check.

        :type kwargs: dict
        :param kwargs: Any keyword arguments to be passed to the check.
        """
        attempts = 0
        successes = 0
        if self.delay_initial_poll:
            time.sleep(self.delay)
        while attempts < self.max_attempts:
            attempts += 1
            if check(*args, **kwargs):
                successes += 1
                if successes >= self.min_successes:
                    return
            else:
                time.sleep(self.delay)
        fail_msg = self._fail_message(attempts, successes)
        raise ConsistencyWaiterException(fail_msg)

    def _fail_message(self, attempts, successes):
        return (
            f'Failed after {attempts} attempts, only had {successes} successes'
        )


class StubbedSession(botocore.session.Session):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_clients = {}
        self._client_stubs = {}

    def create_client(self, service_name, *args, **kwargs):
        if service_name not in self._cached_clients:
            client = self._create_stubbed_client(service_name, *args, **kwargs)
            self._cached_clients[service_name] = client
        return self._cached_clients[service_name]

    def _create_stubbed_client(self, service_name, *args, **kwargs):
        client = super().create_client(service_name, *args, **kwargs)
        stubber = Stubber(client)
        self._client_stubs[service_name] = stubber
        return client

    def stub(self, service_name, *args, **kwargs):
        if service_name not in self._client_stubs:
            self.create_client(service_name, *args, **kwargs)
        return self._client_stubs[service_name]

    def activate_stubs(self):
        for stub in self._client_stubs.values():
            stub.activate()

    def verify_stubs(self):
        for stub in self._client_stubs.values():
            stub.assert_no_pending_responses()


class FreezeTime(ContextDecorator):
    """
    Context manager for mocking out datetime in arbitrary modules when creating
    performing actions like signing which require point in time specificity.

    :type module: module
    :param module: reference to imported module to patch (e.g. botocore.auth.datetime)

    :type date: datetime.datetime
    :param date: datetime object specifying the output for utcnow()
    """

    def __init__(self, module, date=None):
        if date is None:
            date = datetime.datetime.utcnow()
        self.date = date
        self.datetime_patcher = mock.patch.object(
            module, 'datetime', mock.Mock(wraps=datetime.datetime)
        )

    def __enter__(self, *args, **kwargs):
        mock = self.datetime_patcher.start()
        mock.utcnow.return_value = self.date

    def __exit__(self, *args, **kwargs):
        self.datetime_patcher.stop()


def patch_load_service_model(
    session, monkeypatch, service_model_json, ruleset_json
):
    def mock_load_service_model(service_name, type_name, api_version=None):
        if type_name == 'service-2':
            return service_model_json
        if type_name == 'endpoint-rule-set-1':
            return ruleset_json

    loader = session.get_component('data_loader')
    monkeypatch.setattr(loader, 'load_service_model', mock_load_service_model)


def get_checksum_cls(algorithm=DEFAULT_CHECKSUM_ALGORITHM.lower()):
    """
    This pass through is grabbing our internally supported list of checksums
    to ensure we stay in sync, while not exposing them publicly.

    Returns a checksum algorithm class. The default checksum class is used
    if one isn't specified.
    """
    return _CHECKSUM_CLS[algorithm]



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import shutil
import tempfile
import time
from uuid import uuid4

from botocore.exceptions import ClientError
from botocore.session import Session
from tests import BaseEnvVar, mock, random_chars, temporary_file

S3_READ_POLICY_ARN = 'arn:aws:iam::aws:policy/AmazonS3ReadOnlyAccess'


class TestCredentialPrecedence(BaseEnvVar):
    def setUp(self):
        super().setUp()

        # Set the config file to something that doesn't exist so
        # that we don't accidentally load a config.
        os.environ['AWS_CONFIG_FILE'] = '~/.aws/config-missing'

    def create_session(self, *args, **kwargs):
        """
        Create a new session with the given arguments. Additionally,
        this method will set the credentials file to the test credentials
        used by the following test cases.
        """
        kwargs['session_vars'] = {
            'credentials_file': (
                None,
                None,
                os.path.join(os.path.dirname(__file__), 'test-credentials'),
                None,
            )
        }

        return Session(*args, **kwargs)

    def test_access_secret_vs_profile_env(self):
        # If all three are given, then the access/secret keys should
        # take precedence.
        os.environ['AWS_ACCESS_KEY_ID'] = 'env'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'env-secret'
        os.environ['AWS_DEFAULT_PROFILE'] = 'test'

        s = self.create_session()
        credentials = s.get_credentials()

        self.assertEqual(credentials.access_key, 'env')
        self.assertEqual(credentials.secret_key, 'env-secret')

    @mock.patch('botocore.credentials.Credentials')
    def test_access_secret_vs_profile_code(self, credentials_cls):
        # If all three are given, then the access/secret keys should
        # take precedence.
        s = self.create_session(profile='test')
        s.create_client(
            's3', aws_access_key_id='code', aws_secret_access_key='code-secret'
        )

        credentials_cls.assert_called_with(
            access_key='code',
            secret_key='code-secret',
            token=mock.ANY,
            account_id=mock.ANY,
        )

    def test_profile_env_vs_code(self):
        # If the profile is set both by the env var and by code,
        # then the one set by code should take precedence.
        os.environ['AWS_DEFAULT_PROFILE'] = 'test'
        s = self.create_session(profile='default')
        credentials = s.get_credentials()

        self.assertEqual(credentials.access_key, 'default')
        self.assertEqual(credentials.secret_key, 'default-secret')

    @mock.patch('botocore.credentials.Credentials')
    def test_access_secret_env_vs_code(self, credentials_cls):
        # If the access/secret keys are set both as env vars and via
        # code, then those set by code should take precedence.
        os.environ['AWS_ACCESS_KEY_ID'] = 'env'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'secret'
        s = self.create_session()
        s.create_client(
            's3', aws_access_key_id='code', aws_secret_access_key='code-secret'
        )

        credentials_cls.assert_called_with(
            access_key='code',
            secret_key='code-secret',
            token=mock.ANY,
            account_id=mock.ANY,
        )

    def test_access_secret_env_vs_profile_code(self):
        # If access/secret keys are set in the environment, but then a
        # specific profile is passed via code, then the access/secret
        # keys defined in that profile should take precedence over
        # the environment variables. Example:
        #
        # ``aws --profile dev s3 ls``
        #
        os.environ['AWS_ACCESS_KEY_ID'] = 'env'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'env-secret'
        s = self.create_session(profile='test')

        credentials = s.get_credentials()

        self.assertEqual(credentials.access_key, 'test')
        self.assertEqual(credentials.secret_key, 'test-secret')

    def test_honors_aws_shared_credentials_file_env_var(self):
        with temporary_file('w') as f:
            f.write(
                '[default]\n'
                'aws_access_key_id=custom1\n'
                'aws_secret_access_key=custom2\n'
            )
            f.flush()
            os.environ['AWS_SHARED_CREDENTIALS_FILE'] = f.name
            s = Session()
            credentials = s.get_credentials()

            self.assertEqual(credentials.access_key, 'custom1')
            self.assertEqual(credentials.secret_key, 'custom2')


class TestAssumeRoleCredentials(BaseEnvVar):
    def setUp(self):
        self.env_original = os.environ.copy()
        self.environ_copy = os.environ.copy()
        super().setUp()
        os.environ = self.environ_copy
        # The tests rely on manipulating AWS_CONFIG_FILE,
        # but we also need to make sure we don't accidentally
        # pick up the ~/.aws/credentials file either.
        os.environ['AWS_SHARED_CREDENTIALS_FILE'] = str(uuid4())
        self.parent_session = Session()
        self.iam = self.parent_session.create_client('iam')
        self.sts = self.parent_session.create_client('sts')
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, 'config')

        # A role trust policy that allows the current account to call assume
        # role on itself.
        account_id = self.sts.get_caller_identity()['Account']
        self.role_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"AWS": f"arn:aws:iam::{account_id}:root"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.tempdir)
        os.environ = self.env_original.copy()

    def random_name(self):
        return 'botocoretest-' + random_chars(10)

    def create_role(self, policy_document, policy_arn=None):
        name = self.random_name()
        response = self.iam.create_role(
            RoleName=name, AssumeRolePolicyDocument=json.dumps(policy_document)
        )
        self.addCleanup(self.iam.delete_role, RoleName=name)
        if policy_arn:
            self.iam.attach_role_policy(RoleName=name, PolicyArn=policy_arn)
            self.addCleanup(
                self.iam.detach_role_policy,
                RoleName=name,
                PolicyArn=policy_arn,
            )
        return response['Role']

    def create_user(self, policy_arns):
        name = self.random_name()
        user = self.iam.create_user(UserName=name)['User']
        self.addCleanup(self.iam.delete_user, UserName=name)

        for arn in policy_arns:
            self.iam.attach_user_policy(UserName=name, PolicyArn=arn)
            self.addCleanup(
                self.iam.detach_user_policy, UserName=name, PolicyArn=arn
            )

        return user

    def create_creds(self, user_name):
        creds = self.iam.create_access_key(UserName=user_name)['AccessKey']
        self.addCleanup(
            self.iam.delete_access_key,
            UserName=user_name,
            AccessKeyId=creds['AccessKeyId'],
        )
        return creds

    def wait_for_assume_role(
        self,
        role_arn,
        access_key,
        secret_key,
        token=None,
        attempts=30,
        delay=10,
        success_delay=1,
        num_success=4,
    ):
        for _ in range(num_success):
            creds = self._wait_for_assume_role(
                role_arn, access_key, secret_key, token, attempts, delay
            )
            time.sleep(success_delay)
        return creds

    def _wait_for_assume_role(
        self, role_arn, access_key, secret_key, token, attempts, delay
    ):
        # "Why not use the policy simulator?" you might ask. The answer is
        # that the policy simulator will return success far before you can
        # actually make the calls.
        client = self.parent_session.create_client(
            'sts',
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=token,
        )
        attempts_remaining = attempts
        role_session_name = random_chars(10)
        while attempts_remaining > 0:
            attempts_remaining -= 1
            try:
                result = client.assume_role(
                    RoleArn=role_arn, RoleSessionName=role_session_name
                )
                return result['Credentials']
            except ClientError as e:
                code = e.response.get('Error', {}).get('Code')
                if code in ["InvalidClientTokenId", "AccessDenied"]:
                    time.sleep(delay)
                else:
                    raise

        raise Exception(f"Unable to assume role {role_arn}")

    def create_assume_policy(self, role_arn):
        policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Resource": role_arn,
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        name = self.random_name()
        response = self.iam.create_policy(
            PolicyName=name, PolicyDocument=json.dumps(policy_document)
        )
        self.addCleanup(
            self.iam.delete_policy, PolicyArn=response['Policy']['Arn']
        )
        return response['Policy']['Arn']

    def assert_s3_read_only_session(self, session):
        # Calls to S3 should succeed
        s3 = session.create_client('s3')
        s3.list_buckets()

        # Calls to other services should not
        iam = session.create_client('iam')
        try:
            iam.list_groups()
            self.fail("Expected call to list_groups to fail, but it passed.")
        except ClientError as e:
            code = e.response.get('Error', {}).get('Code')
            if code != 'AccessDenied':
                raise

    def test_recursive_assume_role(self):
        # Create the final role, the one that will actually have access to s3
        final_role = self.create_role(self.role_policy, S3_READ_POLICY_ARN)

        # Create the role that can assume the final role
        middle_policy_arn = self.create_assume_policy(final_role['Arn'])
        middle_role = self.create_role(self.role_policy, middle_policy_arn)

        # Create a user that can only assume the middle-man role, and then get
        # static credentials for it.
        user_policy_arn = self.create_assume_policy(middle_role['Arn'])
        user = self.create_user([user_policy_arn])
        user_creds = self.create_creds(user['UserName'])

        # Setup the config file with the profiles we'll be using. For
        # convenience static credentials are placed here instead of putting
        # them in the credentials file.
        config = (
            '[default]\n'
            'aws_access_key_id = %s\n'
            'aws_secret_access_key = %s\n'
            '[profile middle]\n'
            'source_profile = default\n'
            'role_arn = %s\n'
            '[profile final]\n'
            'source_profile = middle\n'
            'role_arn = %s\n'
        )
        config = config % (
            user_creds['AccessKeyId'],
            user_creds['SecretAccessKey'],
            middle_role['Arn'],
            final_role['Arn'],
        )
        with open(self.config_file, 'w') as f:
            f.write(config)

        # Wait for IAM permissions to propagate
        middle_creds = self.wait_for_assume_role(
            role_arn=middle_role['Arn'],
            access_key=user_creds['AccessKeyId'],
            secret_key=user_creds['SecretAccessKey'],
        )
        self.wait_for_assume_role(
            role_arn=final_role['Arn'],
            access_key=middle_creds['AccessKeyId'],
            secret_key=middle_creds['SecretAccessKey'],
            token=middle_creds['SessionToken'],
        )

        # Configure our credentials file to be THE credentials file
        os.environ['AWS_CONFIG_FILE'] = self.config_file

        self.assert_s3_read_only_session(Session(profile='final'))

    def test_assume_role_with_credential_source(self):
        # Create a role with read access to S3
        role = self.create_role(self.role_policy, S3_READ_POLICY_ARN)

        # Create a user that can assume the role and get static credentials
        # for it.
        user_policy_arn = self.create_assume_policy(role['Arn'])
        user = self.create_user([user_policy_arn])
        user_creds = self.create_creds(user['UserName'])

        # Setup the config file with the profile we'll be using.
        config = (
            '[profile assume]\n'
            'role_arn = %s\n'
            'credential_source = Environment\n'
        )
        config = config % role['Arn']
        with open(self.config_file, 'w') as f:
            f.write(config)

        # Wait for IAM permissions to propagate
        self.wait_for_assume_role(
            role_arn=role['Arn'],
            access_key=user_creds['AccessKeyId'],
            secret_key=user_creds['SecretAccessKey'],
        )

        # Setup the environment so that our new config file is THE config
        # file and add the expected credentials since we're using the
        # environment as our credential source.
        os.environ['AWS_CONFIG_FILE'] = self.config_file
        os.environ['AWS_SECRET_ACCESS_KEY'] = user_creds['SecretAccessKey']
        os.environ['AWS_ACCESS_KEY_ID'] = user_creds['AccessKeyId']

        self.assert_s3_read_only_session(Session(profile='assume'))



# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
from io import BytesIO

from dateutil.tz import tzutc
from urllib3.exceptions import ProtocolError as URLLib3ProtocolError
from urllib3.exceptions import ReadTimeoutError as URLLib3ReadTimeoutError

import botocore
from botocore import response
from botocore.awsrequest import AWSResponse
from botocore.exceptions import (
    IncompleteReadError,
    ReadTimeoutError,
    ResponseStreamingError,
)
from tests import unittest
from tests.unit import BaseResponseTest

XMLBODY1 = (
    b'<?xml version="1.0" encoding="UTF-8"?><Error>'
    b'<Code>AccessDenied</Code>'
    b'<Message>Access Denied</Message>'
    b'<RequestId>XXXXXXXXXXXXXXXX</RequestId>'
    b'<HostId>AAAAAAAAAAAAAAAAAAA</HostId>'
    b'</Error>'
)

XMLBODY2 = (
    b'<?xml version="1.0" encoding="UTF-8"?>'
    b'<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
    b'<Name>mybucket</Name><Prefix></Prefix><Marker></Marker>'
    b'<MaxKeys>1000</MaxKeys><IsTruncated>false</IsTruncated>'
    b'<Contents><Key>test.png</Key><LastModified>2014-03-01T17:06:40.000Z</LastModified>'
    b'<ETag>&quot;00000000000000000000000000000000&quot;</ETag><Size>6702</Size>'
    b'<Owner><ID>AAAAAAAAAAAAAAAAAAA</ID>'
    b'<DisplayName>dummy</DisplayName></Owner>'
    b'<StorageClass>STANDARD</StorageClass></Contents></ListBucketResult>'
)


class TestStreamWrapper(unittest.TestCase):
    def assert_lines(self, line_iterator, expected_lines):
        for expected_line in expected_lines:
            self.assertEqual(
                next(line_iterator),
                expected_line,
            )
        # We should have exhausted the iterator.
        with self.assertRaises(StopIteration):
            next(line_iterator)

    def test_streaming_wrapper_validates_content_length(self):
        body = BytesIO(b'1234567890')
        stream = response.StreamingBody(body, content_length=10)
        self.assertEqual(stream.read(), b'1234567890')

    def test_streaming_body_with_invalid_length(self):
        body = BytesIO(b'123456789')
        stream = response.StreamingBody(body, content_length=10)
        with self.assertRaises(IncompleteReadError):
            self.assertEqual(stream.read(9), b'123456789')
            # The next read will have nothing returned and raise
            # an IncompleteReadError because we were expectd 10 bytes, not 9.
            stream.read()

    def test_streaming_body_readable(self):
        body = BytesIO(b'1234567890')
        stream = response.StreamingBody(body, content_length=10)
        self.assertTrue(stream.readable())
        stream.close()
        with self.assertRaises(ValueError):
            stream.readable()

    def test_streaming_body_with_zero_read(self):
        body = BytesIO(b'1234567890')
        stream = response.StreamingBody(body, content_length=10)
        chunk = stream.read(0)
        self.assertEqual(chunk, b'')
        self.assertEqual(stream.read(), b'1234567890')

    def test_streaming_body_with_single_read(self):
        body = BytesIO(b'123456789')
        stream = response.StreamingBody(body, content_length=10)
        with self.assertRaises(IncompleteReadError):
            stream.read()

    def test_streaming_body_readline(self):
        body = BytesIO(b'1234567890\n1234567\n12345\n')
        stream = response.StreamingBody(body, content_length=25)
        chunk = stream.readline()
        self.assertEqual(chunk, b'1234567890\n')
        chunk = stream.readline()
        self.assertEqual(chunk, b'1234567\n')

    def test_streaming_body_readlines(self):
        body = BytesIO(b'1234567890\n1234567890\n12345')
        stream = response.StreamingBody(body, content_length=27)
        chunks = [b'1234567890\n', b'1234567890\n', b'12345']
        self.assertEqual(stream.readlines(), chunks)

    def test_streaming_body_readinto(self):
        body = BytesIO(b"123456789")
        stream = response.StreamingBody(body, content_length=9)
        chunk = bytearray(b"\x00\x00\x00\x00\x00")
        self.assertEqual(5, stream.readinto(chunk))
        self.assertEqual(chunk, bytearray(b"\x31\x32\x33\x34\x35"))
        self.assertEqual(4, stream.readinto(chunk))
        self.assertEqual(chunk, bytearray(b"\x36\x37\x38\x39\x35"))

    def test_streaming_body_readinto_with_invalid_length(self):
        body = BytesIO(b"12")
        stream = response.StreamingBody(body, content_length=9)
        chunk = bytearray(b"\xde\xad\xbe\xef")
        self.assertEqual(2, stream.readinto(chunk))
        self.assertEqual(chunk, bytearray(b"\x31\x32\xbe\xef"))
        with self.assertRaises(IncompleteReadError):
            stream.readinto(chunk)

    def test_streaming_body_readinto_with_empty_buffer(self):
        body = BytesIO(b"12")
        stream = response.StreamingBody(body, content_length=9)
        chunk = bytearray(b"")
        # Does not raise IncompleteReadError
        self.assertEqual(0, stream.readinto(chunk))

    def test_streaming_body_readinto_catches_urllib3_read_timeout(self):
        class TimeoutBody:
            def readinto(*args, **kwargs):
                raise URLLib3ReadTimeoutError(None, None, None)

            def geturl(*args, **kwargs):
                return "http://example.com"

        stream = response.StreamingBody(TimeoutBody(), content_length=None)
        with self.assertRaises(ReadTimeoutError):
            chunk = bytearray(b"\x00\x00\x00\x00\x00")
            stream.readinto(chunk)

    def test_streaming_body_readinto_catches_urllib3_protocol_error(self):
        class ProtocolErrorBody:
            def readinto(*args, **kwargs):
                raise URLLib3ProtocolError(None, None, None)

            def geturl(*args, **kwargs):
                return "http://example.com"

        stream = response.StreamingBody(
            ProtocolErrorBody(), content_length=None
        )
        with self.assertRaises(ResponseStreamingError):
            chunk = bytearray(b"\x00\x00\x00\x00\x00")
            stream.readinto(chunk)

    def test_streaming_body_tell(self):
        body = BytesIO(b'1234567890')
        stream = response.StreamingBody(body, content_length=10)
        self.assertEqual(stream.tell(), 0)
        stream.read(5)
        self.assertEqual(stream.tell(), 5)

    def test_streaming_body_closes(self):
        body = BytesIO(b'1234567890')
        stream = response.StreamingBody(body, content_length=10)
        self.assertFalse(body.closed)
        stream.close()
        self.assertTrue(body.closed)

    def test_default_iter_behavior(self):
        body = BytesIO(b'a' * 2048)
        stream = response.StreamingBody(body, content_length=2048)
        chunks = list(stream)
        self.assertEqual(len(chunks), 2)
        self.assertEqual(chunks, [b'a' * 1024, b'a' * 1024])

    def test_streaming_body_is_an_iterator(self):
        body = BytesIO(b'a' * 1024 + b'b' * 1024 + b'c' * 2)
        stream = response.StreamingBody(body, content_length=2050)
        self.assertEqual(b'a' * 1024, next(stream))
        self.assertEqual(b'b' * 1024, next(stream))
        self.assertEqual(b'c' * 2, next(stream))
        with self.assertRaises(StopIteration):
            next(stream)

    def test_iter_chunks_single_byte(self):
        body = BytesIO(b'abcde')
        stream = response.StreamingBody(body, content_length=5)
        chunks = list(stream.iter_chunks(chunk_size=1))
        self.assertEqual(chunks, [b'a', b'b', b'c', b'd', b'e'])

    def test_iter_chunks_with_leftover(self):
        body = BytesIO(b'abcde')
        stream = response.StreamingBody(body, content_length=5)
        chunks = list(stream.iter_chunks(chunk_size=2))
        self.assertEqual(chunks, [b'ab', b'cd', b'e'])

    def test_iter_chunks_single_chunk(self):
        body = BytesIO(b'abcde')
        stream = response.StreamingBody(body, content_length=5)
        chunks = list(stream.iter_chunks(chunk_size=1024))
        self.assertEqual(chunks, [b'abcde'])

    def test_streaming_line_iterator(self):
        body = BytesIO(b'1234567890\n1234567890\n12345')
        stream = response.StreamingBody(body, content_length=27)
        self.assert_lines(
            stream.iter_lines(),
            [b'1234567890', b'1234567890', b'12345'],
        )

    def test_streaming_line_iterator_ends_newline(self):
        body = BytesIO(b'1234567890\n1234567890\n12345\n')
        stream = response.StreamingBody(body, content_length=28)
        self.assert_lines(
            stream.iter_lines(),
            [b'1234567890', b'1234567890', b'12345'],
        )

    def test_streaming_line_iter_chunk_sizes(self):
        for chunk_size in range(1, 30):
            body = BytesIO(b'1234567890\n1234567890\n12345')
            stream = response.StreamingBody(body, content_length=27)
            self.assert_lines(
                stream.iter_lines(chunk_size),
                [b'1234567890', b'1234567890', b'12345'],
            )

    def test_streaming_line_iterator_keepends(self):
        body = BytesIO(b'1234567890\n1234567890\n12345')
        stream = response.StreamingBody(body, content_length=27)
        self.assert_lines(
            stream.iter_lines(keepends=True),
            [b'1234567890\n', b'1234567890\n', b'12345'],
        )

    def test_catches_urllib3_read_timeout(self):
        class TimeoutBody:
            def read(*args, **kwargs):
                raise URLLib3ReadTimeoutError(None, None, None)

            def geturl(*args, **kwargs):
                return 'http://example.com'

        stream = response.StreamingBody(TimeoutBody(), content_length=None)
        with self.assertRaises(ReadTimeoutError):
            stream.read()

    def test_catches_urllib3_protocol_error(self):
        class ProtocolErrorBody:
            def read(*args, **kwargs):
                raise URLLib3ProtocolError(None, None, None)

            def geturl(*args, **kwargs):
                return 'http://example.com'

        stream = response.StreamingBody(
            ProtocolErrorBody(), content_length=None
        )
        with self.assertRaises(ResponseStreamingError):
            stream.read()

    def test_streaming_line_abstruse_newline_standard(self):
        for chunk_size in range(1, 30):
            body = BytesIO(b'1234567890\r\n1234567890\r\n12345\r\n')
            stream = response.StreamingBody(body, content_length=31)
            self.assert_lines(
                stream.iter_lines(chunk_size),
                [b'1234567890', b'1234567890', b'12345'],
            )

    def test_streaming_line_empty_body(self):
        stream = response.StreamingBody(
            BytesIO(b''),
            content_length=0,
        )
        self.assert_lines(stream.iter_lines(), [])

    def test_streaming_body_as_context_manager(self):
        body = BytesIO(b'1234567890')
        with response.StreamingBody(body, content_length=10) as stream:
            self.assertEqual(stream.read(), b'1234567890')
            self.assertFalse(body.closed)
        self.assertTrue(body.closed)


class FakeRawResponse(BytesIO):
    def stream(self, amt=1024, decode_content=None):
        while True:
            chunk = self.read(amt)
            if not chunk:
                break
            yield chunk


class TestGetResponse(BaseResponseTest):
    maxDiff = None

    def test_get_response_streaming_ok(self):
        headers = {
            'content-type': 'image/png',
            'server': 'AmazonS3',
            'AcceptRanges': 'bytes',
            'transfer-encoding': 'chunked',
            'ETag': '"00000000000000000000000000000000"',
        }
        raw = FakeRawResponse(b'\x89PNG\r\n\x1a\n\x00\x00')

        http_response = AWSResponse(None, 200, headers, raw)

        session = botocore.session.get_session()
        service_model = session.get_service_model('s3')
        operation_model = service_model.operation_model('GetObject')

        res = response.get_response(operation_model, http_response)
        self.assertTrue(isinstance(res[1]['Body'], response.StreamingBody))
        self.assertEqual(res[1]['ETag'], '"00000000000000000000000000000000"')

    def test_get_response_streaming_ng(self):
        headers = {
            'content-type': 'application/xml',
            'date': 'Sat, 08 Mar 2014 12:05:44 GMT',
            'server': 'AmazonS3',
            'transfer-encoding': 'chunked',
            'x-amz-id-2': 'AAAAAAAAAAAAAAAAAAA',
            'x-amz-request-id': 'XXXXXXXXXXXXXXXX',
        }
        raw = FakeRawResponse(XMLBODY1)
        http_response = AWSResponse(None, 403, headers, raw)

        session = botocore.session.get_session()
        service_model = session.get_service_model('s3')
        operation_model = service_model.operation_model('GetObject')

        self.assert_response_with_subset_metadata(
            response.get_response(operation_model, http_response)[1],
            {
                'Error': {'Message': 'Access Denied', 'Code': 'AccessDenied'},
                'ResponseMetadata': {
                    'HostId': 'AAAAAAAAAAAAAAAAAAA',
                    'RequestId': 'XXXXXXXXXXXXXXXX',
                    'HTTPStatusCode': 403,
                },
            },
        )

    def test_get_response_nonstreaming_ok(self):
        headers = {
            'content-type': 'application/xml',
            'date': 'Sun, 09 Mar 2014 02:55:43 GMT',
            'server': 'AmazonS3',
            'transfer-encoding': 'chunked',
            'x-amz-id-2': 'AAAAAAAAAAAAAAAAAAA',
            'x-amz-request-id': 'XXXXXXXXXXXXXXXX',
        }
        raw = FakeRawResponse(XMLBODY1)
        http_response = AWSResponse(None, 403, headers, raw)

        session = botocore.session.get_session()
        service_model = session.get_service_model('s3')
        operation_model = service_model.operation_model('ListObjects')

        self.assert_response_with_subset_metadata(
            response.get_response(operation_model, http_response)[1],
            {
                'ResponseMetadata': {
                    'RequestId': 'XXXXXXXXXXXXXXXX',
                    'HostId': 'AAAAAAAAAAAAAAAAAAA',
                    'HTTPStatusCode': 403,
                },
                'Error': {'Message': 'Access Denied', 'Code': 'AccessDenied'},
            },
        )

    def test_get_response_nonstreaming_ng(self):
        headers = {
            'content-type': 'application/xml',
            'date': 'Sat, 08 Mar 2014 12:05:44 GMT',
            'server': 'AmazonS3',
            'transfer-encoding': 'chunked',
            'x-amz-id-2': 'AAAAAAAAAAAAAAAAAAA',
            'x-amz-request-id': 'XXXXXXXXXXXXXXXX',
        }
        raw = FakeRawResponse(XMLBODY2)
        http_response = AWSResponse(None, 200, headers, raw)

        session = botocore.session.get_session()
        service_model = session.get_service_model('s3')
        operation_model = service_model.operation_model('ListObjects')

        self.assert_response_with_subset_metadata(
            response.get_response(operation_model, http_response)[1],
            {
                'Contents': [
                    {
                        'ETag': '"00000000000000000000000000000000"',
                        'Key': 'test.png',
                        'LastModified': datetime.datetime(
                            2014, 3, 1, 17, 6, 40, tzinfo=tzutc()
                        ),
                        'Owner': {
                            'DisplayName': 'dummy',
                            'ID': 'AAAAAAAAAAAAAAAAAAA',
                        },
                        'Size': 6702,
                        'StorageClass': 'STANDARD',
                    }
                ],
                'IsTruncated': False,
                'Marker': "",
                'MaxKeys': 1000,
                'Name': 'mybucket',
                'Prefix': "",
                'ResponseMetadata': {
                    'RequestId': 'XXXXXXXXXXXXXXXX',
                    'HostId': 'AAAAAAAAAAAAAAAAAAA',
                    'HTTPStatusCode': 200,
                },
            },
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.docs.docstring import (
    ClientMethodDocstring,
    LazyLoadedDocstring,
    PaginatorDocstring,
    WaiterDocstring,
)
from tests import mock, unittest


class MockedLazyLoadedDocstring(LazyLoadedDocstring):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mocked_writer_method = mock.Mock()

    def _write_docstring(self, *args, **kwargs):
        self.mocked_writer_method(*args, **kwargs)


class TestLazyLoadedDocstring(unittest.TestCase):
    def test_raises_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            str(LazyLoadedDocstring())

    def test_expandtabs(self):
        docstring = MockedLazyLoadedDocstring()
        docstring.mocked_writer_method.side_effect = (
            lambda section: section.write('foo\t')
        )
        self.assertEqual('foo ', docstring.expandtabs(1))

    def test_str(self):
        docstring = MockedLazyLoadedDocstring()
        docstring.mocked_writer_method.side_effect = (
            lambda section: section.write('foo')
        )
        self.assertEqual('foo', str(docstring))

    def test_repr(self):
        docstring = MockedLazyLoadedDocstring()
        docstring.mocked_writer_method.side_effect = (
            lambda section: section.write('foo')
        )
        self.assertEqual('foo', repr(docstring))

    def test_is_lazy_loaded(self):
        docstring = MockedLazyLoadedDocstring()
        str(docstring)
        str(docstring)
        # The mock.ANY represents the DocumentStructure that is filled out.
        docstring.mocked_writer_method.assert_called_once_with(mock.ANY)

    def test_args_kwargs_passed(self):
        args = ['foo', 'bar']
        kwargs = {'biz': 'baz'}
        docstring = MockedLazyLoadedDocstring(*args, **kwargs)
        str(docstring)
        # The mock.ANY represents the DocumentStructure that is filled out.
        docstring.mocked_writer_method.assert_called_with(
            mock.ANY, *args, **kwargs
        )


class TestClientMethodDocstring(unittest.TestCase):
    def test_use_correct_docstring_writer(self):
        with mock.patch(
            'botocore.docs.docstring.document_model_driven_method'
        ) as mock_writer:
            docstring = ClientMethodDocstring()
            str(docstring)
            self.assertTrue(mock_writer.called)


class TestWaiterDocstring(unittest.TestCase):
    def test_use_correct_docstring_writer(self):
        with mock.patch(
            'botocore.docs.docstring.document_wait_method'
        ) as mock_writer:
            docstring = WaiterDocstring()
            str(docstring)
            self.assertTrue(mock_writer.called)


class TestPaginatorDocstring(unittest.TestCase):
    def test_use_correct_docstring_writer(self):
        with mock.patch(
            'botocore.docs.docstring.document_paginate_method'
        ) as mock_writer:
            docstring = PaginatorDocstring()
            str(docstring)
            self.assertTrue(mock_writer.called)



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from docutils import nodes
from sphinx.locale import admonitionlabels
from sphinx.writers.html5 import HTML5Translator as SphinxHTML5Translator


class BotoHTML5Translator(SphinxHTML5Translator):
    """Extension of Sphinx's ``HTML5Translator`` for Botocore documentation."""

    IGNORE_IMPLICIT_HEADINGS = [
        '[REQUIRED]',
    ]

    def visit_admonition(self, node, name=""):
        """Uses the h3 tag for admonition titles instead of the p tag."""
        self.body.append(
            self.starttag(node, "div", CLASS=("admonition " + name))
        )
        if name:
            title = (
                f"<h3 class='admonition-title'> {admonitionlabels[name]}</h3>"
            )
            self.body.append(title)

    def is_implicit_heading(self, node):
        """Determines if a node is an implicit heading.

        An implicit heading is represented by a paragraph node whose only
        child is a strong node with text that isnt in `IGNORE_IMPLICIT_HEADINGS`.
        """
        return (
            len(node) == 1
            and isinstance(node[0], nodes.strong)
            and len(node[0]) == 1
            and isinstance(node[0][0], nodes.Text)
            and node[0][0].astext() not in self.IGNORE_IMPLICIT_HEADINGS
        )

    def visit_paragraph(self, node):
        """Visit a paragraph HTML element.

        Replaces implicit headings with an h3 tag and defers to default
        behavior for normal paragraph elements.
        """
        if self.is_implicit_heading(node):
            text = node[0][0]
            self.body.append(f'<h3>{text}</h3>\n')
            # Do not visit the current nodes children or call its depart method.
            raise nodes.SkipNode
        else:
            super().visit_paragraph(node)



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This is a dummy implementation of a credential provider process."""

import argparse
import json


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--raise-error',
        action='store_true',
        help=(
            'If set, this will cause the process to return a non-zero exit code '
            'and print to stderr.'
        ),
    )
    args = parser.parse_args()
    if args.raise_error:
        raise Exception('Failed to fetch credentials.')
    print(
        json.dumps(
            {'AccessKeyId': 'spam', 'SecretAccessKey': 'eggs', 'Version': 1}
        )
    )


if __name__ == "__main__":
    main()



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

import botocore
from botocore.config import Config
from tests import ClientHTTPStubber, patch_load_service_model
from tests.functional.test_useragent import (
    get_captured_ua_strings,
    parse_registered_feature_ids,
)

MOCK_SERVICE_MODEL = {
    "version": "2.0",
    "metadata": {
        "apiVersion": "2020-02-02",
        "endpointPrefix": "mock-bearer-service",
        "jsonVersion": "1.1",
        "protocol": "rest-json",
        "protocols": ["rest-json"],
        "serviceFullName": "Mock Bearer Service",
        "serviceId": "Bearer Service",
        "signatureVersion": "v4",
        "signingName": "bearer-service",
        "uid": "bearer-service-2020-02-02",
        "auth": ["aws.auth#sigv4", "smithy.api#httpBearerAuth"],
    },
    "operations": {
        "MockOperation": {
            "name": "MockOperation",
            "http": {"method": "GET", "requestUri": "/"},
            "input": {"shape": "MockOperationRequest"},
            "documentation": "",
        },
    },
    "shapes": {
        "MockOpParam": {
            "type": "string",
        },
        "MockOperationRequest": {
            "type": "structure",
            "required": ["MockOpParam"],
            "members": {
                "MockOpParam": {
                    "shape": "MockOpParam",
                    "documentation": "",
                    "location": "uri",
                    "locationName": "param",
                },
            },
        },
    },
}

MOCK_RULESET = {
    "version": "1.0",
    "parameters": {},
    "rules": [
        {
            "conditions": [],
            "endpoint": {
                "url": "https://bearer-service.us-west-2.amazonaws.com/"
            },
            "type": "endpoint",
        },
    ],
}


def mocked_supported_services():
    return {"bearer-service"}


@pytest.fixture(autouse=True)
def patch_supported_services(monkeypatch):
    monkeypatch.setattr(
        "botocore.handlers.get_bearer_auth_supported_services",
        mocked_supported_services,
    )


@pytest.mark.parametrize(
    "env_vars",
    [
        {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
        {
            "AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token",
            "AWS_AUTH_SCHEME_PREFERENCE": "sigv4,httpBearerAuth",
        },
    ],
    ids=["valid_token", "token_overrides_auth_scheme_preference"],
)
def test_service_uses_bearer_auth(monkeypatch, patched_session, env_vars):
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)

    patch_load_service_model(
        patched_session, monkeypatch, MOCK_SERVICE_MODEL, MOCK_RULESET
    )

    client = patched_session.create_client(
        "bearer-service",
        region_name="us-west-2",
    )
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.mock_operation(MockOpParam="mock-op-param-value")
        request = http_stubber.requests[0]
        auth_header = request.headers.get("Authorization")
        assert auth_header == b"Bearer bearer-service-token"


@pytest.mark.parametrize(
    "env_vars, config",
    [
        ({}, None),
        ({"AWS_BEARER_TOKEN_FOO_SERVICE": "foo-service-token"}, None),
        (
            {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
            Config(signature_version="v4"),
        ),
        (
            {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
            Config(auth_scheme_preference="sigv4"),
        ),
        (
            {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
            Config(signature_version=botocore.UNSIGNED),
        ),
    ],
    ids=[
        "no_token_configured",
        "invalid_token",
        "signature_version_config_override",
        "auth_scheme_preference_config_override",
        "botocore_unsigned_config_override",
    ],
)
def test_service_does_not_use_bearer_auth(
    monkeypatch, patched_session, env_vars, config
):
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)

    patch_load_service_model(
        patched_session, monkeypatch, MOCK_SERVICE_MODEL, MOCK_RULESET
    )
    client = patched_session.create_client(
        "bearer-service",
        region_name="us-west-2",
        config=config,
    )
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.mock_operation(MockOpParam="mock-op-param-value")
        request = http_stubber.requests[0]
        auth_header = request.headers.get("Authorization")
        assert not (auth_header and auth_header.startswith(b"Bearer"))


@pytest.mark.parametrize(
    "env_vars",
    [
        {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
        {
            "AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token",
            "AWS_AUTH_SCHEME_PREFERENCE": "sigv4,httpBearerAuth",
        },
    ],
    ids=["valid_token", "token_overrides_auth_scheme_preference"],
)
def test_user_agent_has_bearer_service_env_vars_feature_id(
    monkeypatch, patched_session, env_vars
):
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)

    patch_load_service_model(
        patched_session, monkeypatch, MOCK_SERVICE_MODEL, MOCK_RULESET
    )

    client = patched_session.create_client(
        "bearer-service",
        region_name="us-west-2",
    )

    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.mock_operation(MockOpParam="mock-op-param-value")

    ua_string = get_captured_ua_strings(http_stubber)[0]
    feature_ids = parse_registered_feature_ids(ua_string)

    assert "3" in feature_ids


@pytest.mark.parametrize(
    "env_vars, config",
    [
        ({"AWS_BEARER_TOKEN_FOO_SERVICE": "foo-service-token"}, None),
        (
            {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
            Config(signature_version="v4"),
        ),
        (
            {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
            Config(auth_scheme_preference="sigv4"),
        ),
        (
            {"AWS_BEARER_TOKEN_BEARER_SERVICE": "bearer-service-token"},
            Config(signature_version=botocore.UNSIGNED),
        ),
    ],
    ids=[
        "invalid_token",
        "signature_version_config_override",
        "auth_scheme_preference_config_override",
        "botocore_unsigned_config_override",
    ],
)
def test_user_agent_does_not_have_bearer_service_env_vars_feature_id(
    monkeypatch, patched_session, env_vars, config
):
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)

    patch_load_service_model(
        patched_session, monkeypatch, MOCK_SERVICE_MODEL, MOCK_RULESET
    )

    client = patched_session.create_client(
        "bearer-service",
        region_name="us-west-2",
        config=config,
    )

    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.mock_operation(MockOpParam="mock-op-param-value")

    ua_string = get_captured_ua_strings(http_stubber)[0]
    feature_ids = parse_registered_feature_ids(ua_string)

    assert "3" not in feature_ids



import pytest

from botocore import model
from botocore.compat import OrderedDict
from tests import unittest


@pytest.mark.parametrize("property_name", ['api_version', 'protocol'])
def test_missing_model_attribute_raises_exception(property_name):
    service_model = model.ServiceModel({'metadata': {'endpointPrefix': 'foo'}})
    with pytest.raises(model.UndefinedModelAttributeError):
        getattr(service_model, property_name)


class TestServiceId(unittest.TestCase):
    def test_hypenize_replaces_spaces(self):
        self.assertEqual(
            model.ServiceId('my service').hyphenize(), 'my-service'
        )

    def test_hyphenize_lower_cases(self):
        self.assertEqual(model.ServiceId('MyService').hyphenize(), 'myservice')


class TestServiceModel(unittest.TestCase):
    def setUp(self):
        self.model = {
            'metadata': {
                'protocol': 'query',
                'endpointPrefix': 'endpoint-prefix',
                'serviceId': 'MyService',
            },
            'documentation': 'Documentation value',
            'operations': {},
            'shapes': {'StringShape': {'type': 'string'}},
        }
        self.error_shapes = {
            'ExceptionOne': {
                'exception': True,
                'type': 'structure',
                'members': {},
            },
            'ExceptionTwo': {
                'exception': True,
                'type': 'structure',
                'members': {},
                'error': {'code': 'FooCode'},
            },
        }
        self.service_model = model.ServiceModel(self.model)

    def test_metadata_available(self):
        # You should be able to access the metadata in a service description
        # through the service model object.
        self.assertEqual(self.service_model.metadata.get('protocol'), 'query')

    def test_service_name_can_be_overriden(self):
        service_model = model.ServiceModel(
            self.model, service_name='myservice'
        )
        self.assertEqual(service_model.service_name, 'myservice')

    def test_service_name_defaults_to_endpoint_prefix(self):
        self.assertEqual(self.service_model.service_name, 'endpoint-prefix')

    def test_service_id(self):
        self.assertEqual(self.service_model.service_id, 'MyService')

    def test_hyphenize_service_id(self):
        self.assertEqual(
            self.service_model.service_id.hyphenize(), 'myservice'
        )

    def test_service_id_does_not_exist(self):
        service_model = {
            'metadata': {
                'protocol': 'query',
                'endpointPrefix': 'endpoint-prefix',
            },
            'documentation': 'Documentation value',
            'operations': {},
            'shapes': {'StringShape': {'type': 'string'}},
        }
        service_name = 'myservice'
        service_model = model.ServiceModel(service_model, service_name)
        with self.assertRaisesRegex(
            model.UndefinedModelAttributeError, service_name
        ):
            service_model.service_id

    def test_operation_does_not_exist(self):
        with self.assertRaises(model.OperationNotFoundError):
            self.service_model.operation_model('NoExistOperation')

    def test_signing_name_defaults_to_endpoint_prefix(self):
        self.assertEqual(self.service_model.signing_name, 'endpoint-prefix')

    def test_documentation_exposed_as_property(self):
        self.assertEqual(
            self.service_model.documentation, 'Documentation value'
        )

    def test_shape_names(self):
        self.assertEqual(self.service_model.shape_names, ['StringShape'])

    def test_repr_has_service_name(self):
        self.assertEqual(
            repr(self.service_model), 'ServiceModel(endpoint-prefix)'
        )

    def test_shape_for_error_code(self):
        self.model['shapes'].update(self.error_shapes)
        self.service_model = model.ServiceModel(self.model)
        shape = self.service_model.shape_for_error_code('ExceptionOne')
        self.assertEqual(shape.name, 'ExceptionOne')
        shape = self.service_model.shape_for_error_code('FooCode')
        self.assertEqual(shape.name, 'ExceptionTwo')

    def test_error_shapes(self):
        self.model['shapes'].update(self.error_shapes)
        self.service_model = model.ServiceModel(self.model)
        error_shapes = self.service_model.error_shapes
        error_shape_names = [shape.name for shape in error_shapes]
        self.assertEqual(len(error_shape_names), 2)
        self.assertIn('ExceptionOne', error_shape_names)
        self.assertIn('ExceptionTwo', error_shape_names)

    def test_client_context_params(self):
        service_model = model.ServiceModel(
            {
                'metadata': {
                    'protocol': 'query',
                    'endpointPrefix': 'endpoint-prefix',
                    'serviceId': 'MyServiceWithClientContextParams',
                },
                'documentation': 'Documentation value',
                'operations': {},
                'shapes': {},
                'clientContextParams': {
                    'stringClientContextParam': {
                        'type': 'string',
                        'documentation': 'str-valued',
                    },
                    'booleanClientContextParam': {
                        'type': 'boolean',
                        'documentation': 'bool-valued',
                    },
                },
            }
        )
        self.assertEqual(len(service_model.client_context_parameters), 2)
        client_ctx_param1 = service_model.client_context_parameters[0]
        client_ctx_param2 = service_model.client_context_parameters[1]
        self.assertEqual(client_ctx_param1.name, 'stringClientContextParam')
        self.assertEqual(client_ctx_param1.type, 'string')
        self.assertEqual(client_ctx_param1.documentation, 'str-valued')
        self.assertEqual(client_ctx_param2.name, 'booleanClientContextParam')

    def test_client_context_params_absent(self):
        self.assertIsInstance(
            self.service_model.client_context_parameters, list
        )
        self.assertEqual(len(self.service_model.client_context_parameters), 0)

    def test_protocol_resolution_without_protocols_trait(self):
        self.assertIsNone(self.service_model.metadata.get('protocols'))
        self.assertEqual(self.service_model.resolved_protocol, 'query')

    def test_protocol_resolution_picks_highest_supported(self):
        self.service_model.metadata['protocols'] = ['query', 'json']
        self.assertEqual(self.service_model.resolved_protocol, 'json')

    def test_protocol_raises_error_for_unsupported_protocol(self):
        self.service_model.metadata['protocols'] = ['wrongprotocol']
        with self.assertRaises(model.UnsupportedServiceProtocolsError):
            self.service_model.resolved_protocol


class TestOperationModelFromService(unittest.TestCase):
    def setUp(self):
        self.model = {
            'metadata': {'protocol': 'query', 'endpointPrefix': 'foo'},
            'documentation': '',
            'operations': {
                'OperationName': {
                    'http': {
                        'method': 'POST',
                        'requestUri': '/',
                    },
                    'name': 'OperationName',
                    'input': {'shape': 'OperationNameRequest'},
                    'output': {
                        'shape': 'OperationNameResponse',
                    },
                    'errors': [{'shape': 'NoSuchResourceException'}],
                    'documentation': 'Docs for OperationName',
                    'authtype': 'v4',
                    'auth': ['aws.auth#sigv4'],
                },
                'OperationTwo': {
                    'http': {
                        'method': 'POST',
                        'requestUri': '/',
                    },
                    'name': 'OperationTwo',
                    'input': {'shape': 'OperationNameRequest'},
                    'output': {
                        'shape': 'OperationNameResponse',
                    },
                    'errors': [{'shape': 'NoSuchResourceException'}],
                    'documentation': 'Docs for OperationTwo',
                },
                'PayloadOperation': {
                    'http': {
                        'method': 'POST',
                        'requestUri': '/',
                    },
                    'name': 'PayloadOperation',
                    'input': {'shape': 'PayloadOperationRequest'},
                    'output': {
                        'shape': 'PayloadOperationResponse',
                    },
                    'errors': [{'shape': 'NoSuchResourceException'}],
                    'documentation': 'Docs for PayloadOperation',
                    'requestcompression': {'encodings': ['gzip']},
                },
                'NoBodyOperation': {
                    'http': {
                        'method': 'GET',
                        'requestUri': '/',
                    },
                    'name': 'NoBodyOperation',
                    'input': {'shape': 'NoBodyOperationRequest'},
                    'output': {
                        'shape': 'OperationNameResponse',
                    },
                    'errors': [{'shape': 'NoSuchResourceException'}],
                    'documentation': 'Docs for NoBodyOperation',
                },
                'ContextParamOperation': {
                    'http': {
                        'method': 'POST',
                        'requestUri': '/',
                    },
                    'name': 'ContextParamOperation',
                    'input': {'shape': 'ContextParamOperationRequest'},
                    'output': {'shape': 'OperationNameResponse'},
                    'errors': [{'shape': 'NoSuchResourceException'}],
                    'documentation': 'Docs for ContextParamOperation',
                    'staticContextParams': {
                        'stringStaticContextParam': {
                            'value': 'Static Context Param Value',
                        },
                        'booleanStaticContextParam': {
                            'value': True,
                        },
                    },
                },
            },
            'shapes': {
                'OperationNameRequest': {
                    'type': 'structure',
                    'members': {
                        'Arg1': {'shape': 'stringType'},
                        'Arg2': {'shape': 'stringType'},
                    },
                },
                'OperationNameResponse': {
                    'type': 'structure',
                    'members': {
                        'String': {
                            'shape': 'stringType',
                        }
                    },
                },
                'PayloadOperationRequest': {
                    'type': 'structure',
                    'members': {
                        'Arg1': {'shape': 'TestConfig'},
                        'Arg2': {'shape': 'stringType'},
                    },
                    'payload': 'Arg1',
                },
                'PayloadOperationResponse': {
                    'type': 'structure',
                    'members': {
                        'String': {
                            'shape': 'stringType',
                        }
                    },
                    'payload': 'String',
                },
                'NoBodyOperationRequest': {
                    'type': 'structure',
                    'members': {
                        'data': {
                            'location': 'header',
                            'locationName': 'x-amz-data',
                            'shape': 'stringType',
                        }
                    },
                },
                'ContextParamOperationRequest': {
                    'type': 'structure',
                    'members': {
                        'ContextParamArg': {
                            'shape': 'stringType',
                            'contextParam': {
                                'name': 'contextParamName',
                            },
                        }
                    },
                    'payload': 'ContextParamArg',
                },
                'NoSuchResourceException': {
                    'type': 'structure',
                    'members': {},
                },
                'stringType': {
                    'type': 'string',
                },
                'TestConfig': {
                    'type': 'structure',
                    'members': {'timeout': {'shape': 'stringType'}},
                },
            },
        }
        self.service_model = model.ServiceModel(self.model)

    def test_wire_name_always_matches_model(self):
        service_model = model.ServiceModel(self.model)
        operation = model.OperationModel(
            self.model['operations']['OperationName'], service_model, 'Foo'
        )
        self.assertEqual(operation.name, 'Foo')
        self.assertEqual(operation.wire_name, 'OperationName')

    def test_operation_name_in_repr(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertIn('OperationName', repr(operation))

    def test_name_and_wire_name_defaults_to_same_value(self):
        service_model = model.ServiceModel(self.model)
        operation = model.OperationModel(
            self.model['operations']['OperationName'], service_model
        )
        self.assertEqual(operation.name, 'OperationName')
        self.assertEqual(operation.wire_name, 'OperationName')

    def test_name_from_service(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertEqual(operation.name, 'OperationName')

    def test_name_from_service_model_when_differs_from_name(self):
        self.model['operations']['Foo'] = self.model['operations'][
            'OperationName'
        ]
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('Foo')
        self.assertEqual(operation.name, 'Foo')

    def test_operation_input_model(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertEqual(operation.name, 'OperationName')
        # Operations should also have a reference to the top level metadata.
        self.assertEqual(operation.metadata['protocol'], 'query')
        self.assertEqual(operation.http['method'], 'POST')
        self.assertEqual(operation.http['requestUri'], '/')
        shape = operation.input_shape
        self.assertEqual(shape.name, 'OperationNameRequest')
        self.assertEqual(list(sorted(shape.members)), ['Arg1', 'Arg2'])

    def test_has_documentation_property(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertEqual(operation.documentation, 'Docs for OperationName')

    def test_service_model_available_from_operation_model(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        # This is an identity comparison because we don't implement
        # __eq__, so we may need to change this in the future.
        self.assertEqual(operation.service_model, service_model)

    def test_operation_output_model(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        output = operation.output_shape
        self.assertEqual(list(output.members), ['String'])
        self.assertFalse(operation.has_streaming_output)

    def test_operation_shape_not_required(self):
        # It's ok if there's no output shape. We'll just get a return value of
        # None.
        del self.model['operations']['OperationName']['output']
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        output_shape = operation.output_shape
        self.assertIsNone(output_shape)

    def test_error_shapes(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        # OperationName only has a NoSuchResourceException
        self.assertEqual(len(operation.error_shapes), 1)
        self.assertEqual(
            operation.error_shapes[0].name, 'NoSuchResourceException'
        )

    def test_has_auth(self):
        operation = self.service_model.operation_model('OperationName')
        self.assertEqual(operation.auth, ["aws.auth#sigv4"])

    def test_auth_not_set(self):
        operation = self.service_model.operation_model('OperationTwo')
        self.assertIsNone(operation.auth)

    def test_has_resolved_auth_type(self):
        operation = self.service_model.operation_model('OperationName')
        self.assertEqual(operation.resolved_auth_type, 'v4')

    def test_resolved_auth_type_not_set(self):
        operation = self.service_model.operation_model('OperationTwo')
        self.assertIsNone(operation.resolved_auth_type)

    def test_has_auth_type(self):
        operation = self.service_model.operation_model('OperationName')
        self.assertEqual(operation.auth_type, 'v4')

    def test_auth_type_not_set(self):
        operation = self.service_model.operation_model('OperationTwo')
        self.assertIsNone(operation.auth_type)

    def test_deprecated_present(self):
        self.model['operations']['OperationName']['deprecated'] = True
        service_model = model.ServiceModel(self.model)
        operation_name = service_model.operation_model('OperationName')
        self.assertTrue(operation_name.deprecated)

    def test_deprecated_present_false(self):
        self.model['operations']['OperationName']['deprecated'] = False
        service_model = model.ServiceModel(self.model)
        operation_name = service_model.operation_model('OperationName')
        self.assertFalse(operation_name.deprecated)

    def test_deprecated_absent(self):
        service_model = model.ServiceModel(self.model)
        operation_two = service_model.operation_model('OperationTwo')
        self.assertFalse(operation_two.deprecated)

    def test_endpoint_operation_present(self):
        self.model['operations']['OperationName']['endpointoperation'] = True
        service_model = model.ServiceModel(self.model)
        operation_name = service_model.operation_model('OperationName')
        self.assertTrue(operation_name.is_endpoint_discovery_operation)

    def test_endpoint_operation_present_false(self):
        self.model['operations']['OperationName']['endpointoperation'] = False
        service_model = model.ServiceModel(self.model)
        operation_name = service_model.operation_model('OperationName')
        self.assertFalse(operation_name.is_endpoint_discovery_operation)

    def test_endpoint_operation_absent(self):
        operation_two = self.service_model.operation_model('OperationName')
        self.assertFalse(operation_two.is_endpoint_discovery_operation)

    def test_endpoint_discovery_required(self):
        operation = self.model['operations']['OperationName']
        operation['endpointdiscovery'] = {'required': True}
        service_model = model.ServiceModel(self.model)
        self.assertTrue(service_model.endpoint_discovery_required)

    def test_endpoint_discovery_required_false(self):
        self.model['operations']['OperationName']['endpointdiscovery'] = {}
        service_model = model.ServiceModel(self.model)
        self.assertFalse(service_model.endpoint_discovery_required)

    def test_endpoint_discovery_required_no_value(self):
        operation = self.model['operations']['OperationName']
        self.assertTrue(operation.get('endpointdiscovery') is None)
        service_model = model.ServiceModel(self.model)
        self.assertFalse(service_model.endpoint_discovery_required)

    def test_endpoint_discovery_present(self):
        operation = self.model['operations']['OperationName']
        operation['endpointdiscovery'] = {'required': True}
        service_model = model.ServiceModel(self.model)
        operation_name = service_model.operation_model('OperationName')
        self.assertTrue(operation_name.endpoint_discovery.get('required'))

    def test_endpoint_discovery_absent(self):
        operation_name = self.service_model.operation_model('OperationName')
        self.assertIsNone(operation_name.endpoint_discovery)

    def test_http_checksum_absent(self):
        operation_name = self.service_model.operation_model('OperationName')
        self.assertEqual(operation_name.http_checksum, {})

    def test_http_checksum_present(self):
        operation = self.model['operations']['OperationName']
        operation['httpChecksum'] = {
            "requestChecksumRequired": True,
            "requestAlgorithmMember": "ChecksumAlgorithm",
            "requestValidationModeMember": "ChecksumMode",
            "responseAlgorithms": ["crc32", "crc32c", "sha256", "sha1"],
        }
        service_model = model.ServiceModel(self.model)
        operation_model = service_model.operation_model('OperationName')
        http_checksum = operation_model.http_checksum
        self.assertEqual(
            http_checksum["requestChecksumRequired"],
            True,
        )
        self.assertEqual(
            http_checksum["requestAlgorithmMember"],
            "ChecksumAlgorithm",
        )
        self.assertEqual(
            http_checksum["requestValidationModeMember"],
            "ChecksumMode",
        )
        self.assertEqual(
            http_checksum["responseAlgorithms"],
            ["crc32", "crc32c", "sha256", "sha1"],
        )

    def test_context_parameter_present(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('ContextParamOperation')
        self.assertEqual(len(operation.context_parameters), 1)
        context_param = operation.context_parameters[0]
        self.assertEqual(context_param.name, 'contextParamName')
        self.assertEqual(context_param.member_name, 'ContextParamArg')

    def test_context_parameter_absent(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationTwo')
        self.assertIsInstance(operation.context_parameters, list)
        self.assertEqual(len(operation.context_parameters), 0)

    def test_static_context_parameter_present(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('ContextParamOperation')
        self.assertEqual(len(operation.static_context_parameters), 2)
        static_ctx_param1 = operation.static_context_parameters[0]
        static_ctx_param2 = operation.static_context_parameters[1]
        self.assertEqual(static_ctx_param1.name, 'stringStaticContextParam')
        self.assertEqual(static_ctx_param1.value, 'Static Context Param Value')
        self.assertEqual(static_ctx_param2.name, 'booleanStaticContextParam')
        self.assertEqual(static_ctx_param2.value, True)

    def test_static_context_parameter_absent(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationTwo')
        self.assertIsInstance(operation.static_context_parameters, list)
        self.assertEqual(len(operation.static_context_parameters), 0)

    def test_request_compression(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('PayloadOperation')
        self.assertEqual(
            operation.request_compression, {'encodings': ['gzip']}
        )


class TestOperationModelEventStreamTypes(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.model = {
            'metadata': {'protocol': 'rest-xml', 'endpointPrefix': 'foo'},
            'documentation': '',
            'operations': {
                'OperationName': {
                    'http': {
                        'method': 'POST',
                        'requestUri': '/',
                    },
                    'name': 'OperationName',
                    'input': {'shape': 'OperationRequest'},
                    'output': {'shape': 'OperationResponse'},
                }
            },
            'shapes': {
                'NormalStructure': {
                    'type': 'structure',
                    'members': {'Input': {'shape': 'StringType'}},
                },
                'OperationRequest': {
                    'type': 'structure',
                    'members': {
                        'String': {'shape': 'StringType'},
                        "Body": {'shape': 'EventStreamStructure'},
                    },
                    'payload': 'Body',
                },
                'OperationResponse': {
                    'type': 'structure',
                    'members': {
                        'String': {'shape': 'StringType'},
                        "Body": {'shape': 'EventStreamStructure'},
                    },
                    'payload': 'Body',
                },
                'StringType': {'type': 'string'},
                'BlobType': {'type': 'blob'},
                'EventStreamStructure': {
                    'eventstream': True,
                    'type': 'structure',
                    'members': {
                        'EventA': {'shape': 'EventAStructure'},
                        'EventB': {'shape': 'EventBStructure'},
                    },
                },
                'EventAStructure': {
                    'event': True,
                    'type': 'structure',
                    'members': {
                        'Payload': {'shape': 'BlobType', 'eventpayload': True},
                        'Header': {'shape': 'StringType', 'eventheader': True},
                    },
                },
                'EventBStructure': {
                    'event': True,
                    'type': 'structure',
                    'members': {'Records': {'shape': 'StringType'}},
                },
            },
        }

    def update_operation(self, **kwargs):
        operation = self.model['operations']['OperationName']
        operation.update(kwargs)

    def test_event_stream_input_for_operation(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertTrue(operation.has_event_stream_input)
        event_stream_input = operation.get_event_stream_input()
        self.assertEqual(event_stream_input.name, 'EventStreamStructure')

    def test_no_event_stream_input_for_operation(self):
        self.update_operation(input={'shape': 'NormalStructure'})
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertFalse(operation.has_event_stream_input)
        self.assertEqual(operation.get_event_stream_input(), None)

    def test_event_stream_output_for_operation(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertTrue(operation.has_event_stream_output)
        output = operation.get_event_stream_output()
        self.assertEqual(output.name, 'EventStreamStructure')

    def test_no_event_stream_output_for_operation(self):
        self.update_operation(output={'shape': 'NormalStructure'})
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertFalse(operation.has_event_stream_output)
        self.assertEqual(operation.get_event_stream_output(), None)

    def test_no_output_shape(self):
        self.update_operation(output=None)
        del self.model['operations']['OperationName']['output']
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertFalse(operation.has_event_stream_output)
        self.assertEqual(operation.get_event_stream_output(), None)


class TestOperationModelStreamingTypes(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.model = {
            'metadata': {'protocol': 'query', 'endpointPrefix': 'foo'},
            'documentation': '',
            'operations': {
                'OperationName': {
                    'name': 'OperationName',
                    'input': {
                        'shape': 'OperationRequest',
                    },
                    'output': {
                        'shape': 'OperationResponse',
                    },
                }
            },
            'shapes': {
                'OperationRequest': {
                    'type': 'structure',
                    'members': {
                        'String': {
                            'shape': 'stringType',
                        },
                        "Body": {
                            'shape': 'blobType',
                        },
                    },
                    'payload': 'Body',
                },
                'OperationResponse': {
                    'type': 'structure',
                    'members': {
                        'String': {
                            'shape': 'stringType',
                        },
                        "Body": {
                            'shape': 'blobType',
                        },
                    },
                    'payload': 'Body',
                },
                'stringType': {
                    'type': 'string',
                },
                'blobType': {'type': 'blob'},
            },
        }

    def remove_payload(self, type):
        self.model['shapes']['Operation' + type].pop('payload')

    def test_streaming_input_for_operation(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertTrue(operation.has_streaming_input)
        self.assertEqual(operation.get_streaming_input().name, 'blobType')

    def test_not_streaming_input_for_operation(self):
        self.remove_payload('Request')
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertFalse(operation.has_streaming_input)
        self.assertEqual(operation.get_streaming_input(), None)

    def test_streaming_output_for_operation(self):
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertTrue(operation.has_streaming_output)
        self.assertEqual(operation.get_streaming_output().name, 'blobType')

    def test_not_streaming_output_for_operation(self):
        self.remove_payload('Response')
        service_model = model.ServiceModel(self.model)
        operation = service_model.operation_model('OperationName')
        self.assertFalse(operation.has_streaming_output)
        self.assertEqual(operation.get_streaming_output(), None)


class TestDeepMerge(unittest.TestCase):
    def setUp(self):
        self.shapes = {
            'SetQueueAttributes': {
                'type': 'structure',
                'members': {
                    'MapExample': {
                        'shape': 'StrToStrMap',
                        'locationName': 'Attribute',
                    },
                },
            },
            'SetQueueAttributes2': {
                'type': 'structure',
                'members': {
                    'MapExample': {
                        'shape': 'StrToStrMap',
                        'locationName': 'Attribute2',
                    },
                },
            },
            'StrToStrMap': {
                'type': 'map',
                'key': {'shape': 'StringType', 'locationName': 'Name'},
                'value': {'shape': 'StringType', 'locationName': 'Value'},
                'flattened': True,
                'name': 'NotAttribute',
            },
            'StringType': {'type': 'string'},
        }
        self.shape_resolver = model.ShapeResolver(self.shapes)

    def test_deep_merge(self):
        shape = self.shape_resolver.get_shape_by_name('SetQueueAttributes')
        map_merged = shape.members['MapExample']
        # map_merged has a serialization as a member trait as well as
        # in the StrToStrMap.
        # The member trait should have precedence.
        self.assertEqual(
            map_merged.serialization,
            # member beats the definition.
            {
                'name': 'Attribute',
                # From the definition.
                'flattened': True,
            },
        )
        # Ensure we don't merge/mutate the original dicts.
        self.assertEqual(map_merged.key.serialization['name'], 'Name')
        self.assertEqual(map_merged.value.serialization['name'], 'Value')
        self.assertEqual(map_merged.key.serialization['name'], 'Name')

    def test_merges_copy_dict(self):
        shape = self.shape_resolver.get_shape_by_name('SetQueueAttributes')
        map_merged = shape.members['MapExample']
        self.assertEqual(map_merged.serialization.get('name'), 'Attribute')

        shape2 = self.shape_resolver.get_shape_by_name('SetQueueAttributes2')
        map_merged2 = shape2.members['MapExample']
        self.assertEqual(map_merged2.serialization.get('name'), 'Attribute2')


class TestShapeResolver(unittest.TestCase):
    def test_get_shape_by_name(self):
        shape_map = {
            'Foo': {
                'type': 'structure',
                'members': {
                    'Bar': {'shape': 'StringType'},
                    'Baz': {'shape': 'StringType'},
                },
            },
            "StringType": {"type": "string"},
        }
        resolver = model.ShapeResolver(shape_map)
        shape = resolver.get_shape_by_name('Foo')
        self.assertEqual(shape.name, 'Foo')
        self.assertEqual(shape.type_name, 'structure')

    def test_resolve_shape_reference(self):
        shape_map = {
            'Foo': {
                'type': 'structure',
                'members': {
                    'Bar': {'shape': 'StringType'},
                    'Baz': {'shape': 'StringType'},
                },
            },
            "StringType": {"type": "string"},
        }
        resolver = model.ShapeResolver(shape_map)
        shape = resolver.resolve_shape_ref({'shape': 'StringType'})
        self.assertEqual(shape.name, 'StringType')
        self.assertEqual(shape.type_name, 'string')

    def test_resolve_shape_references_with_member_traits(self):
        shape_map = {
            'Foo': {
                'type': 'structure',
                'members': {
                    'Bar': {'shape': 'StringType'},
                    'Baz': {'shape': 'StringType', 'locationName': 'other'},
                },
            },
            "StringType": {"type": "string"},
        }
        resolver = model.ShapeResolver(shape_map)
        shape = resolver.resolve_shape_ref(
            {'shape': 'StringType', 'locationName': 'other'}
        )
        self.assertEqual(shape.serialization['name'], 'other')
        self.assertEqual(shape.name, 'StringType')

    def test_serialization_cache(self):
        shape_map = {
            'Foo': {
                'type': 'structure',
                'members': {
                    'Baz': {'shape': 'StringType', 'locationName': 'other'},
                },
            },
            "StringType": {"type": "string"},
        }
        resolver = model.ShapeResolver(shape_map)
        shape = resolver.resolve_shape_ref(
            {'shape': 'StringType', 'locationName': 'other'}
        )
        self.assertEqual(shape.serialization['name'], 'other')
        # serialization is computed on demand, and a cache is kept.
        # This is just verifying that trying to access serialization again
        # gives the same result.  We don't actually care that it's cached,
        # we just care that the cache doesn't mess with correctness.
        self.assertEqual(shape.serialization['name'], 'other')

    def test_shape_overrides(self):
        shape_map = {
            "StringType": {
                "type": "string",
                "documentation": "Original documentation",
            }
        }
        resolver = model.ShapeResolver(shape_map)
        shape = resolver.get_shape_by_name('StringType')
        self.assertEqual(shape.documentation, 'Original documentation')

        shape = resolver.resolve_shape_ref(
            {'shape': 'StringType', 'documentation': 'override'}
        )
        self.assertEqual(shape.documentation, 'override')

    def test_shape_type_structure(self):
        shapes = {
            'ChangePasswordRequest': {
                'type': 'structure',
                'members': {
                    'OldPassword': {'shape': 'passwordType'},
                    'NewPassword': {'shape': 'passwordType'},
                },
            },
            'passwordType': {
                "type": "string",
            },
        }
        resolver = model.ShapeResolver(shapes)
        shape = resolver.get_shape_by_name('ChangePasswordRequest')
        self.assertEqual(shape.type_name, 'structure')
        self.assertEqual(shape.name, 'ChangePasswordRequest')
        self.assertEqual(
            list(sorted(shape.members)), ['NewPassword', 'OldPassword']
        )
        self.assertEqual(shape.members['OldPassword'].name, 'passwordType')
        self.assertEqual(shape.members['OldPassword'].type_name, 'string')
        self.assertEqual(shape.error_code, None)

    def test_exception_error_code(self):
        shapes = {
            'FooException': {
                'exception': True,
                'type': 'structure',
                'members': {},
            }
        }
        # Test without explicit error code
        resolver = model.ShapeResolver(shapes)
        shape = resolver.get_shape_by_name('FooException')
        self.assertTrue(shape.metadata['exception'])
        self.assertEqual(shape.error_code, 'FooException')
        # Test with explicit error code
        shapes['FooException']['error'] = {'code': 'ExceptionCode'}
        resolver = model.ShapeResolver(shapes)
        shape = resolver.get_shape_by_name('FooException')
        self.assertTrue(shape.metadata['exception'])
        self.assertEqual(shape.error_code, 'ExceptionCode')

    def test_shape_metadata(self):
        shapes = {
            'ChangePasswordRequest': {
                'type': 'structure',
                'required': ['OldPassword', 'NewPassword'],
                'members': {
                    'OldPassword': {'shape': 'passwordType'},
                    'NewPassword': {'shape': 'passwordType'},
                },
            },
            'passwordType': {
                "type": "string",
                "min": 1,
                "max": 128,
                "pattern": ".*",
                "sensitive": True,
            },
        }
        resolver = model.ShapeResolver(shapes)
        shape = resolver.get_shape_by_name('ChangePasswordRequest')
        self.assertEqual(
            shape.metadata['required'], ['OldPassword', 'NewPassword']
        )
        member = shape.members['OldPassword']
        self.assertEqual(member.metadata['min'], 1)
        self.assertEqual(member.metadata['max'], 128)
        self.assertEqual(member.metadata['pattern'], '.*')
        self.assertEqual(member.metadata['sensitive'], True)

    def test_error_shape_metadata(self):
        shapes = {
            'ResourceNotFoundException': {
                'type': 'structure',
                'members': {
                    'message': {
                        'shape': 'ErrorMessage',
                    }
                },
                'exception': True,
                'retryable': {'throttling': True},
            }
        }
        resolver = model.ShapeResolver(shapes)
        shape = resolver.get_shape_by_name('ResourceNotFoundException')
        self.assertEqual(
            shape.metadata,
            {'exception': True, 'retryable': {'throttling': True}},
        )

    def test_shape_list(self):
        shapes = {
            'mfaDeviceListType': {
                "type": "list",
                "member": {"shape": "MFADevice"},
            },
            'MFADevice': {
                'type': 'structure',
                'members': {'UserName': {'shape': 'userNameType'}},
            },
            'userNameType': {'type': 'string'},
        }
        resolver = model.ShapeResolver(shapes)
        shape = resolver.get_shape_by_name('mfaDeviceListType')
        self.assertEqual(shape.member.type_name, 'structure')
        self.assertEqual(shape.member.name, 'MFADevice')
        self.assertEqual(list(shape.member.members), ['UserName'])

    def test_shape_does_not_exist(self):
        resolver = model.ShapeResolver({})
        with self.assertRaises(model.NoShapeFoundError):
            resolver.get_shape_by_name('NoExistShape')

    def test_missing_type_key(self):
        shapes = {'UnknownType': {'NotTheTypeKey': 'someUnknownType'}}
        resolver = model.ShapeResolver(shapes)
        with self.assertRaises(model.InvalidShapeError):
            resolver.get_shape_by_name('UnknownType')

    def test_bad_shape_ref(self):
        # This is an example of a denormalized model,
        # which should raise an exception.
        shapes = {
            'Struct': {
                'type': 'structure',
                'members': {
                    'A': {'type': 'string'},
                    'B': {'type': 'string'},
                },
            }
        }
        resolver = model.ShapeResolver(shapes)
        with self.assertRaises(model.InvalidShapeReferenceError):
            struct = resolver.get_shape_by_name('Struct')
            # Resolving the members will fail because
            # the 'A' and 'B' members are not shape refs.
            struct.members

    def test_shape_name_in_repr(self):
        shapes = {
            'StringType': {
                'type': 'string',
            }
        }
        resolver = model.ShapeResolver(shapes)
        self.assertIn(
            'StringType', repr(resolver.get_shape_by_name('StringType'))
        )


class TestBuilders(unittest.TestCase):
    def test_structure_shape_builder_with_scalar_types(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {'type': 'string'},
                'B': {'type': 'integer'},
            }
        ).build_model()
        self.assertIsInstance(shape, model.StructureShape)
        self.assertEqual(sorted(list(shape.members)), ['A', 'B'])
        self.assertEqual(shape.members['A'].type_name, 'string')
        self.assertEqual(shape.members['B'].type_name, 'integer')

    def test_structure_shape_with_structure_type(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {
                    'type': 'structure',
                    'members': {
                        'A-1': {'type': 'string'},
                    },
                },
            }
        ).build_model()
        self.assertIsInstance(shape, model.StructureShape)
        self.assertEqual(list(shape.members), ['A'])
        self.assertEqual(shape.members['A'].type_name, 'structure')
        self.assertEqual(list(shape.members['A'].members), ['A-1'])

    def test_structure_shape_with_list(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {'type': 'list', 'member': {'type': 'string'}},
            }
        ).build_model()
        self.assertIsInstance(shape.members['A'], model.ListShape)
        self.assertEqual(shape.members['A'].member.type_name, 'string')

    def test_structure_shape_with_map_type(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {
                    'type': 'map',
                    'key': {'type': 'string'},
                    'value': {'type': 'string'},
                }
            }
        ).build_model()
        self.assertIsInstance(shape.members['A'], model.MapShape)
        map_shape = shape.members['A']
        self.assertEqual(map_shape.key.type_name, 'string')
        self.assertEqual(map_shape.value.type_name, 'string')

    def test_nested_structure(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {
                    'type': 'structure',
                    'members': {
                        'B': {
                            'type': 'structure',
                            'members': {
                                'C': {
                                    'type': 'string',
                                }
                            },
                        }
                    },
                }
            }
        ).build_model()
        self.assertEqual(
            shape.members['A'].members['B'].members['C'].type_name, 'string'
        )

    def test_enum_values_on_string_used(self):
        b = model.DenormalizedStructureBuilder()
        enum_values = ['foo', 'bar', 'baz']
        shape = b.with_members(
            {
                'A': {
                    'type': 'string',
                    'enum': enum_values,
                },
            }
        ).build_model()
        self.assertIsInstance(shape, model.StructureShape)
        string_shape = shape.members['A']
        self.assertIsInstance(string_shape, model.StringShape)
        self.assertEqual(string_shape.metadata['enum'], enum_values)
        self.assertEqual(string_shape.enum, enum_values)

    def test_documentation_on_shape_used(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {
                    'type': 'string',
                    'documentation': 'MyDocs',
                },
            }
        ).build_model()
        self.assertEqual(shape.members['A'].documentation, 'MyDocs')

    def test_min_max_used_in_metadata(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {
                    'type': 'string',
                    'documentation': 'MyDocs',
                    'min': 2,
                    'max': 3,
                },
            }
        ).build_model()
        metadata = shape.members['A'].metadata
        self.assertEqual(metadata.get('min'), 2)
        self.assertEqual(metadata.get('max'), 3)

    def test_use_shape_name_when_provided(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            {
                'A': {
                    'type': 'string',
                    'shape_name': 'MyStringShape',
                },
            }
        ).build_model()
        self.assertEqual(shape.members['A'].name, 'MyStringShape')

    def test_unknown_shape_type(self):
        b = model.DenormalizedStructureBuilder()
        with self.assertRaises(model.InvalidShapeError):
            b.with_members(
                {
                    'A': {
                        'type': 'brand-new-shape-type',
                    },
                }
            ).build_model()

    def test_ordered_shape_builder(self):
        b = model.DenormalizedStructureBuilder()
        shape = b.with_members(
            OrderedDict(
                [
                    ('A', {'type': 'string'}),
                    (
                        'B',
                        {
                            'type': 'structure',
                            'members': OrderedDict(
                                [
                                    ('C', {'type': 'string'}),
                                    ('D', {'type': 'string'}),
                                ]
                            ),
                        },
                    ),
                ]
            )
        ).build_model()

        # Members should be in order
        self.assertEqual(['A', 'B'], list(shape.members.keys()))

        # Nested structure members should *also* stay ordered
        self.assertEqual(['C', 'D'], list(shape.members['B'].members.keys()))


if __name__ == '__main__':
    unittest.main()



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import re
from datetime import datetime

from botocore.config import Config
from botocore.stub import Stubber
from tests import (
    BaseSessionTest,
    ClientHTTPStubber,
    assert_url_equal,
    mock,
    temporary_file,
)

_V4_SIGNING_REGION_REGEX = re.compile(
    r'AWS4-HMAC-SHA256 Credential=\w+/\d+/(?P<signing_region>[a-z0-9-]+)/'
)


class TestSTSPresignedUrl(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.client = self.session.create_client('sts', 'us-west-2')
        # Makes sure that no requests will go through
        self.stubber = Stubber(self.client)
        self.stubber.activate()

    def test_presigned_url_contains_no_content_type(self):
        timestamp = datetime(2017, 3, 22, 0, 0)
        with mock.patch('botocore.auth.datetime.datetime') as _datetime:
            _datetime.utcnow.return_value = timestamp
            url = self.client.generate_presigned_url('get_caller_identity', {})

        # There should be no 'content-type' in x-amz-signedheaders
        expected_url = (
            'https://sts.amazonaws.com/?Action=GetCallerIdentity&'
            'Version=2011-06-15&X-Amz-Algorithm=AWS4-HMAC-SHA256&'
            'X-Amz-Credential=access_key%2F20170322%2Fus-east-1%2Fsts%2F'
            'aws4_request&X-Amz-Date=20170322T000000Z&X-Amz-Expires=3600&'
            'X-Amz-SignedHeaders=host&X-Amz-Signature=767845d2ee858069a598d5f'
            '8b497b75c7d57356885b1b3dba46dbbc0fc62bf5a'
        )
        assert_url_equal(url, expected_url)


class TestSTSEndpoints(BaseSessionTest):
    def create_sts_client(
        self, region, endpoint_url=None, use_ssl=True, config=None
    ):
        return self.session.create_client(
            'sts',
            region_name=region,
            endpoint_url=endpoint_url,
            use_ssl=use_ssl,
            config=config,
        )

    def set_sts_regional_for_config_file(self, fileobj, config_val):
        fileobj.write(f'[default]\nsts_regional_endpoints={config_val}\n')
        fileobj.flush()
        self.environ['AWS_CONFIG_FILE'] = fileobj.name

    def assert_request_sent(
        self, sts, expected_url, expected_signing_region=None
    ):
        body = (
            b'<GetCallerIdentityResponse>'
            b'  <GetCallerIdentityResult>'
            b'     <Arn>arn:aws:iam::123456789012:user/myuser</Arn>'
            b'     <UserId>UserID</UserId>'
            b'     <Account>123456789012</Account>'
            b'   </GetCallerIdentityResult>'
            b'   <ResponseMetadata>'
            b'     <RequestId>some-request</RequestId>'
            b'   </ResponseMetadata>'
            b'</GetCallerIdentityResponse>'
        )
        with ClientHTTPStubber(sts) as http_stubber:
            http_stubber.add_response(body=body)
            sts.get_caller_identity()
            captured_request = http_stubber.requests[0]
            self.assertEqual(captured_request.url, expected_url)
            if expected_signing_region:
                self.assertEqual(
                    self._get_signing_region(captured_request),
                    expected_signing_region,
                )

    def _get_signing_region(self, request):
        authorization_val = request.headers['Authorization'].decode('utf-8')
        match = _V4_SIGNING_REGION_REGEX.match(authorization_val)
        return match.group('signing_region')

    def test_legacy_region_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('us-west-2')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.amazonaws.com/',
            expected_signing_region='us-east-1',
        )

    def test_legacy_region_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('us-west-2')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.us-west-2.amazonaws.com/',
            expected_signing_region='us-west-2',
        )

    def test_fips_endpoint_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('us-west-2-fips')
        self.assert_request_sent(
            sts,
            expected_url='https://sts-fips.us-west-2.amazonaws.com/',
            expected_signing_region='us-west-2',
        )

    def test_fips_endpoint_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('us-west-2-fips')
        self.assert_request_sent(
            sts,
            expected_url='https://sts-fips.us-west-2.amazonaws.com/',
            expected_signing_region='us-west-2',
        )

    def test_dualstack_endpoint_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        dualstack_config = Config(use_dualstack_endpoint=True)
        sts = self.create_sts_client('us-west-2', config=dualstack_config)
        self.assert_request_sent(
            sts,
            expected_url='https://sts.us-west-2.api.aws/',
            expected_signing_region='us-west-2',
        )

    def test_dualstack_endpoint_with_regional_configured(self):
        dualstack_config = Config(use_dualstack_endpoint=True)
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('us-west-2', config=dualstack_config)
        self.assert_request_sent(
            sts,
            expected_url='https://sts.us-west-2.api.aws/',
            expected_signing_region='us-west-2',
        )

    def test_nonlegacy_region_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('ap-east-1')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.ap-east-1.amazonaws.com/',
            expected_signing_region='ap-east-1',
        )

    def test_nonlegacy_region_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('ap-east-1')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.ap-east-1.amazonaws.com/',
            expected_signing_region='ap-east-1',
        )

    def test_nonaws_partition_region_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('cn-north-1')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.cn-north-1.amazonaws.com.cn/',
            expected_signing_region='cn-north-1',
        )

    def test_nonaws_partition_region_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('cn-north-1')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.cn-north-1.amazonaws.com.cn/',
            expected_signing_region='cn-north-1',
        )

    def test_global_region_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('aws-global')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.amazonaws.com/',
            expected_signing_region='us-east-1',
        )

    def test_global_region_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('aws-global')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.amazonaws.com/',
            expected_signing_region='us-east-1',
        )

    def test_defaults_to_global_endpoint_for_legacy_region(self):
        sts = self.create_sts_client('us-west-2')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.amazonaws.com/',
            expected_signing_region='us-east-1',
        )

    def test_defaults_to_regional_endpoint_for_nonlegacy_region(self):
        sts = self.create_sts_client('ap-east-1')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.ap-east-1.amazonaws.com/',
            expected_signing_region='ap-east-1',
        )

    def test_configure_sts_regional_from_config_file(self):
        with temporary_file('w') as f:
            self.set_sts_regional_for_config_file(f, 'regional')
            sts = self.create_sts_client('us-west-2')
            self.assert_request_sent(
                sts,
                expected_url='https://sts.us-west-2.amazonaws.com/',
            )

    def test_env_var_overrides_config_file(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        with temporary_file('w') as f:
            self.set_sts_regional_for_config_file(f, 'regional')
            sts = self.create_sts_client('us-west-2')
            self.assert_request_sent(
                sts, expected_url='https://sts.amazonaws.com/'
            )

    def test_user_provided_endpoint_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client(
            'us-west-2', endpoint_url='https://custom.com'
        )
        self.assert_request_sent(sts, expected_url='https://custom.com/')

    def test_user_provided_endpoint_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client(
            'us-west-2', endpoint_url='https://custom.com'
        )
        self.assert_request_sent(sts, expected_url='https://custom.com/')

    def test_http_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('us-west-2', use_ssl=False)
        self.assert_request_sent(sts, expected_url='http://sts.amazonaws.com/')

    def test_client_for_unknown_region(self):
        sts = self.create_sts_client('not-real')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.not-real.amazonaws.com/',
            expected_signing_region='not-real',
        )

    def test_client_for_unknown_region_with_legacy_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'legacy'
        sts = self.create_sts_client('not-real')
        self.assert_request_sent(
            sts, expected_url='https://sts.not-real.amazonaws.com/'
        )

    def test_client_for_unknown_region_with_regional_configured(self):
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        sts = self.create_sts_client('not-real')
        self.assert_request_sent(
            sts,
            expected_url='https://sts.not-real.amazonaws.com/',
            expected_signing_region='not-real',
        )


def test_assume_role_with_saml_no_region_custom_endpoint(patched_session):
    # When an endpoint_url and no region are given, AssumeRoleWithSAML should
    # resolve to the endpoint_url and succeed, not fail in endpoint resolution:
    # https://github.com/aws/aws-cli/issues/7455

    client = patched_session.create_client(
        'sts', region_name=None, endpoint_url="https://custom.endpoint.aws"
    )
    assert client.meta.region_name is None

    mock_response_body = b"""\
<AssumeRoleWithSAMLResponse xmlns="https://sts.amazonaws.com/doc/2011-06-15/">
    <AssumeRoleWithSAMLResult></AssumeRoleWithSAMLResult>
</AssumeRoleWithSAMLResponse>
"""
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response(body=mock_response_body)
        client.assume_role_with_saml(
            RoleArn='arn:aws:iam::123456789:role/RoleA',
            PrincipalArn='arn:aws:iam::123456789:role/RoleB',
            SAMLAssertion='xxxx',
        )
    captured_request = http_stubber.requests[0]
    assert captured_request.url == "https://custom.endpoint.aws/"



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from botocore.context import (
    ClientContext,
    get_context,
    reset_context,
    set_context,
    start_as_current_context,
    with_current_context,
)


@pytest.fixture
def custom_context():
    ctx = ClientContext()
    ctx.features.add('CUSTOM')
    return ctx


class TestContext:
    def test_get_context_returns_none(self):
        ctx = get_context()
        assert ctx is None

    def test_get_context_returns_current(self, client_context):
        ctx = get_context()
        assert ctx == client_context

    def test_set_context(self, custom_context):
        token = set_context(custom_context)
        ctx = get_context()
        assert ctx == custom_context
        reset_context(token)

    def test_start_as_current_context(self):
        with start_as_current_context():
            ctx = get_context()
            ctx.features.add('FOO')
            assert ctx.features == {'FOO'}
        ctx = get_context()
        assert ctx is None

    def test_nested_start_as_current_context(self):
        with start_as_current_context():
            ctx = get_context()
            ctx.features.add('FOO')
            with start_as_current_context():
                ctx = get_context()
                ctx.features.add('BAR')
                assert ctx.features == {'FOO', 'BAR'}
            ctx = get_context()
            assert ctx.features == {'FOO'}
        ctx = get_context()
        assert ctx is None

    def test_start_as_current_context_with_param(self, custom_context):
        with start_as_current_context(custom_context):
            ctx = get_context()
            assert ctx.features == {'CUSTOM'}

    def test_with_current_context(self):
        @with_current_context()
        def do_something():
            ctx = get_context()
            assert ctx is not None

        do_something()

    def test_with_current_context_with_hook(self):
        def register_fake():
            ctx = get_context()
            ctx.features.add('FOO')

        @with_current_context(register_fake)
        def do_something():
            ctx = get_context()
            assert ctx.features == {'FOO'}

        do_something()



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.docs.client import (
    ClientContextParamsDocumenter,
    ClientDocumenter,
    ClientExceptionsDocumenter,
)
from tests.unit.docs import BaseDocsTest


class TestClientDocumenter(BaseDocsTest):
    def setUp(self):
        super().setUp()
        exception_shape = {
            'SomeException': {
                'exception': True,
                'type': 'structure',
                'members': {'Message': {'shape': 'String'}},
            }
        }
        self.add_shape(exception_shape)
        self.add_shape_to_params('Biz', 'String')
        self.add_shape_to_errors('SomeException')
        self.setup_client()
        self.client_documenter = ClientDocumenter(
            self.client, self.root_services_path
        )

    def test_document_client(self):
        self.client_documenter.document_client(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '======',
                'Client',
                '======',
                '.. py:class:: MyService.Client',
                '  A low-level client representing AWS MyService',
                '  AWS MyService Description',
                '    client = session.create_client(\'myservice\')',
                'These are the available methods:',
                '  myservice/client/can_paginate',
                '  myservice/client/get_paginator',
                '  myservice/client/get_waiter',
                '  myservice/client/sample_operation',
            ]
        )
        self.assert_contains_lines_in_order(
            ['.. py:method:: MyService.Client.can_paginate(operation_name)'],
            self.get_nested_service_contents(
                'myservice', 'client', 'can_paginate'
            ),
        )
        self.assert_contains_lines_in_order(
            ['.. py:method:: MyService.Client.get_paginator(operation_name)'],
            self.get_nested_service_contents(
                'myservice', 'client', 'get_paginator'
            ),
        )
        self.assert_contains_lines_in_order(
            ['.. py:method:: MyService.Client.get_waiter(waiter_name)'],
            self.get_nested_service_contents(
                'myservice', 'client', 'get_waiter'
            ),
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: MyService.Client.sample_operation(**kwargs)',
                '  **Request Syntax**',
                '  ::',
                '    response = client.sample_operation(',
                '        Biz=\'string\'',
                '    )',
                '  :type Biz: string',
                '  :param Biz:',
                '  :rtype: dict',
                '  :returns:',
                '    **Response Syntax**',
                '    ::',
                '      {',
                '          \'Biz\': \'string\'',
                '      }',
                '    **Response Structure**',
                '    - *(dict) --*',
                '      - **Biz** *(string) --*',
                '**Exceptions**',
                '*   :py:class:`MyService.Client.exceptions.SomeException`',
            ],
            self.get_nested_service_contents(
                'myservice', 'client', 'sample_operation'
            ),
        )


class TestClientExceptionsDocumenter(BaseDocsTest):
    def setup_documenter(self):
        self.setup_client()
        self.exceptions_documenter = ClientExceptionsDocumenter(
            self.client, self.root_services_path
        )

    def test_no_modeled_exceptions(self):
        self.setup_documenter()
        self.exceptions_documenter.document_exceptions(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '=================',
                'Client Exceptions',
                '=================',
                'Client exceptions are available',
                'This client has no modeled exception classes.',
            ]
        )

    def test_modeled_exceptions(self):
        exception_shape = {
            'SomeException': {
                'exception': True,
                'type': 'structure',
                'members': {'Message': {'shape': 'String'}},
            }
        }
        self.add_shape(exception_shape)
        self.setup_documenter()
        self.exceptions_documenter.document_exceptions(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '=================',
                'Client Exceptions',
                '=================',
                'Client exceptions are available',
                'The available client exceptions are:',
                '.. toctree::',
                ':maxdepth: 1',
                ':titlesonly:',
                '  myservice/client/exceptions/SomeException',
            ]
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:class:: MyService.Client.exceptions.SomeException',
                '**Example**',
                '::',
                'except client.exceptions.SomeException as e:',
                '.. py:attribute:: response',
                '**Syntax**',
                '{',
                "'Message': 'string',",
                "'Error': {",
                "'Code': 'string',",
                "'Message': 'string'",
                '}',
                '}',
                '**Structure**',
                '- *(dict) --*',
                '- **Message** *(string) --* ',
                '- **Error** *(dict) --* ',
                '- **Code** *(string) --* ',
                '- **Message** *(string) --* ',
            ],
            self.get_nested_service_contents(
                'myservice', 'client/exceptions', 'SomeException'
            ),
        )


class TestClientContextParamsDocumenter(BaseDocsTest):
    def setUp(self):
        super().setUp()
        self.json_model['clientContextParams'] = {
            'ClientContextParam1': {
                'type': 'string',
                'documentation': 'A client context param',
            },
            'ClientContextParam2': {
                'type': 'boolean',
                'documentation': 'A second client context param',
            },
        }
        self.setup_client()
        service_model = self.client.meta.service_model
        self.context_params_documenter = ClientContextParamsDocumenter(
            service_model.service_name, service_model.client_context_parameters
        )

    def test_client_context_params(self):
        self.context_params_documenter.document_context_params(
            self.doc_structure
        )
        self.assert_contains_lines_in_order(
            [
                '========================',
                'Client Context Parameters',
                '========================',
                'Client context parameters are configurable',
                'The available ``myservice`` client context params are:',
                '* ``client_context_param1`` (string) - A client context param',
                '* ``client_context_param2`` (boolean) - A second client context param',
            ]
        )



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

import pytest
from dateutil.tz import tzutc

import botocore.auth
from tests import FreezeTime

HOSTNAME = "peccy.dsql.us-east-1.on.aws"
REGION = "us-east-1"
AWS_ACCESS_KEY_ID = "akid"
AWS_SECRET_ACCESS_KEY = "secret"
DATE = datetime.datetime(2024, 8, 27, tzinfo=tzutc())


@pytest.fixture()
def client(patched_session):
    return patched_session.create_client(
        "dsql",
        REGION,
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
    )


@FreezeTime(botocore.auth.datetime, date=DATE)
def test_generate_db_connect_auth_token(client):
    auth_token = client.generate_db_connect_auth_token(
        Hostname=HOSTNAME, Region=REGION
    )
    assert f"{HOSTNAME}/?Action=DbConnect" in auth_token
    assert (
        f"X-Amz-Credential={AWS_ACCESS_KEY_ID}%2F20240827%2F{REGION}%2Fdsql%2Faws4_request"
        in auth_token
    )
    assert "X-Amz-Expires=900" in auth_token
    # Asserts that there is no scheme in the URL
    assert auth_token.startswith(HOSTNAME)


@FreezeTime(botocore.auth.datetime, date=DATE)
def test_generate_db_connect_admin_auth_token(client):
    auth_token = client.generate_db_connect_admin_auth_token(
        Hostname=HOSTNAME, Region=REGION
    )
    assert f"{HOSTNAME}/?Action=DbConnectAdmin" in auth_token
    assert (
        f"X-Amz-Credential={AWS_ACCESS_KEY_ID}%2F20240827%2F{REGION}%2Fdsql%2Faws4_request"
        in auth_token
    )
    assert "X-Amz-Expires=900" in auth_token
    # Asserts that there is no scheme in the URL
    assert auth_token.startswith(HOSTNAME)


def test_generate_db_connect_auth_token_extra_params(client):
    region = "us-west-2"
    expires_in = 3600
    auth_token = client.generate_db_connect_auth_token(
        HOSTNAME, Region=region, ExpiresIn=expires_in
    )
    assert f"X-Amz-Expires={expires_in}" in auth_token
    assert region in auth_token


def test_generate_db_connect_admin_auth_token_extra_params(client):
    region = "us-west-2"
    expires_in = 3600
    auth_token = client.generate_db_connect_admin_auth_token(
        HOSTNAME, Region=region, ExpiresIn=expires_in
    )
    assert f"&X-Amz-Expires={expires_in}" in auth_token
    assert region in auth_token



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

import pytest
from dateutil.tz import tzutc

import botocore.session
from botocore.auth import S3ExpressAuth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials, RefreshableCredentials
from botocore.utils import S3ExpressIdentityCache
from tests import ClientHTTPStubber, mock

ACCESS_KEY = "AKIDEXAMPLE"
SECRET_KEY = "wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY"
TOKEN = "TOKEN"

EXPRESS_ACCESS_KEY = "EXPRESS_AKIDEXAMPLE"
EXPRESS_SECRET_KEY = "EXPRESS_53cr37"
EXPRESS_TOKEN = "EXPRESS_TOKEN"

CREDENTIALS = Credentials(ACCESS_KEY, SECRET_KEY, TOKEN)
ENDPOINT = "https://s3.us-west-2.amazonaws.com"

S3EXPRESS_BUCKET = "mytestbucket--usw2-az5--x-s3"


DATE = datetime.datetime(2023, 11, 26, 0, 0, 0, tzinfo=tzutc())


CREATE_SESSION_RESPONSE = (
    b"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
    b"<CreateSessionResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">"
    b"<Credentials><SessionToken>EXPRESS_TOKEN</SessionToken>"
    b"<SecretAccessKey>EXPRESS_53cr37</SecretAccessKey>"
    b"<AccessKeyId>EXPRESS_AKIDEXAMPLE</AccessKeyId>"
    b"<Expiration>2023-11-28T01:46:39Z</Expiration>"
    b"</Credentials></CreateSessionResult>"
)


@pytest.fixture
def default_s3_client():
    session = botocore.session.Session()
    return session.create_client(
        's3',
        'us-west-2',
        aws_access_key_id=ACCESS_KEY,
        aws_secret_access_key=SECRET_KEY,
        aws_session_token=TOKEN,
    )


@pytest.fixture
def mock_datetime():
    with mock.patch('datetime.datetime', spec=True) as mock_datetime:
        yield mock_datetime


def _assert_expected_create_session_call(request, bucket):
    assert bucket in request.url
    assert request.url.endswith('?session')
    assert request.method == "GET"


class TestS3ExpressAuth:
    def _assert_has_expected_headers(self, request, header_list=None):
        if header_list is None:
            header_list = ["Authorization", "X-Amz-S3Session-Token"]

        for header in header_list:
            assert header in request.headers

    def test_s3express_auth_requires_instance_cache(self):
        assert hasattr(S3ExpressAuth, "REQUIRES_IDENTITY_CACHE")
        assert S3ExpressAuth.REQUIRES_IDENTITY_CACHE is True

    def test_s3_express_auth_headers(self):
        request = AWSRequest(
            method='GET',
            url=ENDPOINT,
        )
        auth_instance = S3ExpressAuth(
            CREDENTIALS, "s3", "us-west-2", identity_cache={}
        )

        # Confirm there is no existing auth info on request
        assert 'Authorization' not in request.headers
        auth_instance.add_auth(request)
        self._assert_has_expected_headers(request)

        # Confirm we're not including the unsupported X-Amz-Security-Token
        # header for S3Express.
        assert 'X-Amz-Security-Token' not in request.headers


class TestS3ExpressIdentityCache:
    def test_default_s3_express_cache(self, default_s3_client, mock_datetime):
        mock_datetime.now.return_value = DATE
        mock_datetime.utcnow.return_value = DATE

        identity_cache = S3ExpressIdentityCache(
            default_s3_client,
            RefreshableCredentials,
        )
        with ClientHTTPStubber(default_s3_client) as stubber:
            stubber.add_response(body=CREATE_SESSION_RESPONSE)
            credentials = identity_cache.get_credentials('my_bucket')

            assert credentials.access_key == EXPRESS_ACCESS_KEY
            assert credentials.secret_key == EXPRESS_SECRET_KEY
            assert credentials.token == EXPRESS_TOKEN

    def test_s3_express_cache_one_network_call(
        self, default_s3_client, mock_datetime
    ):
        mock_datetime.now.return_value = DATE
        mock_datetime.utcnow.return_value = DATE
        bucket = 'my_bucket'

        identity_cache = S3ExpressIdentityCache(
            default_s3_client,
            RefreshableCredentials,
        )
        with ClientHTTPStubber(default_s3_client) as stubber:
            # Only set one response
            stubber.add_response(body=CREATE_SESSION_RESPONSE)

            first_creds = identity_cache.get_credentials(bucket)
            second_creds = identity_cache.get_credentials(bucket)

            # Confirm we got back the same credentials
            assert first_creds == second_creds

            # Confirm we didn't hit the API twice
            assert len(stubber.requests) == 1
            _assert_expected_create_session_call(stubber.requests[0], bucket)

    def test_s3_express_cache_multiple_buckets(
        self, default_s3_client, mock_datetime
    ):
        mock_datetime.now.return_value = DATE
        mock_datetime.utcnow.return_value = DATE
        bucket = 'my_bucket'
        other_bucket = 'other_bucket'

        identity_cache = S3ExpressIdentityCache(
            default_s3_client,
            RefreshableCredentials,
        )
        with ClientHTTPStubber(default_s3_client) as stubber:
            stubber.add_response(body=CREATE_SESSION_RESPONSE)
            stubber.add_response(body=CREATE_SESSION_RESPONSE)

            identity_cache.get_credentials(bucket)
            identity_cache.get_credentials(other_bucket)

            # Confirm we hit the API for each bucket
            assert len(stubber.requests) == 2
            _assert_expected_create_session_call(stubber.requests[0], bucket)
            _assert_expected_create_session_call(
                stubber.requests[1], other_bucket
            )


class TestS3ExpressRequests:
    def _assert_standard_sigv4_signature(self, headers):
        sigv4_auth_preamble = (
            b'AWS4-HMAC-SHA256 Credential='
            b'AKIDEXAMPLE/20231126/us-west-2/s3express/aws4_request'
        )
        assert 'Authorization' in headers
        assert headers['Authorization'].startswith(sigv4_auth_preamble)

    def _assert_s3express_credentials(self, headers):
        s3express_auth_preamble = (
            b'AWS4-HMAC-SHA256 Credential='
            b'EXPRESS_AKIDEXAMPLE/20231126/us-west-2/s3express/aws4_request'
        )
        assert 'Authorization' in headers
        assert 'x-amz-s3session-token' in headers
        assert headers['Authorization'].startswith(s3express_auth_preamble)
        assert headers['x-amz-s3session-token'] == b'EXPRESS_TOKEN'

    def _assert_checksum_algorithm_added(self, algorithm, headers):
        algorithm_header_name = f"x-amz-checksum-{algorithm}"
        assert algorithm_header_name in headers

    def _call_get_object(self, client):
        return client.get_object(
            Bucket=S3EXPRESS_BUCKET,
            Key='my-test-object',
        )

    def test_create_bucket(self, default_s3_client, mock_datetime):
        mock_datetime.utcnow.return_value = DATE

        with ClientHTTPStubber(default_s3_client) as stubber:
            stubber.add_response()

            default_s3_client.create_bucket(
                Bucket=S3EXPRESS_BUCKET,
                CreateBucketConfiguration={
                    'Location': {
                        'Name': 'usw2-az5',
                        'Type': 'AvailabilityZone',
                    },
                    'Bucket': {
                        "DataRedundancy": "SingleAvailabilityZone",
                        "Type": "Directory",
                    },
                },
            )

        # Confirm we don't call CreateSession for create_bucket
        assert len(stubber.requests) == 1
        self._assert_standard_sigv4_signature(stubber.requests[0].headers)

    def test_get_object(self, default_s3_client, mock_datetime):
        mock_datetime.utcnow.return_value = DATE
        mock_datetime.now.return_value = DATE

        with ClientHTTPStubber(default_s3_client) as stubber:
            stubber.add_response(body=CREATE_SESSION_RESPONSE)
            stubber.add_response()

            self._call_get_object(default_s3_client)

        # Confirm we called CreateSession for create_bucket
        assert len(stubber.requests) == 2
        _assert_expected_create_session_call(
            stubber.requests[0], S3EXPRESS_BUCKET
        )
        self._assert_standard_sigv4_signature(stubber.requests[0].headers)

        # Confirm actual PutObject request was signed with Session credentials
        self._assert_s3express_credentials(stubber.requests[1].headers)

    def test_cache_with_multiple_requests(
        self, default_s3_client, mock_datetime
    ):
        mock_datetime.utcnow.return_value = DATE
        mock_datetime.now.return_value = DATE

        with ClientHTTPStubber(default_s3_client) as stubber:
            stubber.add_response(body=CREATE_SESSION_RESPONSE)
            stubber.add_response()
            stubber.add_response()

            self._call_get_object(default_s3_client)
            self._call_get_object(default_s3_client)

        # Confirm we called CreateSession for create_bucket once
        assert len(stubber.requests) == 3
        _assert_expected_create_session_call(
            stubber.requests[0], S3EXPRESS_BUCKET
        )
        self._assert_standard_sigv4_signature(stubber.requests[0].headers)

        # Confirm we signed both called with S3Express credentials
        self._assert_s3express_credentials(stubber.requests[1].headers)
        self._assert_s3express_credentials(stubber.requests[2].headers)

    def test_delete_objects_injects_correct_checksum(
        self, default_s3_client, mock_datetime
    ):
        mock_datetime.utcnow.return_value = DATE
        mock_datetime.now.return_value = DATE

        with ClientHTTPStubber(default_s3_client) as stubber:
            stubber.add_response(body=CREATE_SESSION_RESPONSE)
            stubber.add_response()

            default_s3_client.delete_objects(
                Bucket=S3EXPRESS_BUCKET,
                Delete={
                    "Objects": [
                        {
                            "Key": "my-obj",
                            "VersionId": "1",
                        }
                    ]
                },
            )

        # Confirm we called CreateSession for create_bucket
        assert len(stubber.requests) == 2
        _assert_expected_create_session_call(
            stubber.requests[0], S3EXPRESS_BUCKET
        )
        self._assert_standard_sigv4_signature(stubber.requests[0].headers)

        # Confirm we signed both called with S3Express credentials
        self._assert_s3express_credentials(stubber.requests[1].headers)
        self._assert_checksum_algorithm_added(
            'crc32', stubber.requests[1].headers
        )



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

import botocore.session


def test_ambiguous_error_parsing():
    # We map errors from services to modeled errors based on the error code, but
    # cloudwatch has two errors that are both modeled to the `ResourceNotFound`
    # code: `DashboardNotFoundError` and `ResourceNotFound`.  Botocore picks the one
    # that is defined later, which in this case is `ResourceNotFound`.  This test
    # ensures that we continue to select the latter error going forward.
    session = botocore.session.get_session()
    cloudwatch = session.create_client('cloudwatch', region_name='us-west-2')
    with pytest.raises(cloudwatch.exceptions.ResourceNotFound) as exception:
        cloudwatch.get_dashboard(
            DashboardName='dashboard-which-does-not-exist'
        )

    error_response = exception.value.response['Error']
    assert error_response['Type'] == 'Sender'
    assert error_response['Code'] == 'ResourceNotFound'
    assert (
        exception.value.response['ResponseMetadata']['HTTPStatusCode'] == 404
    )
